import pytest

import numpy as np

from ase.units import Bohr, Ha

from gpaw import GPAW
from gpaw.sphere.integrate import integrate_lebedev

from gpaw.response import ResponseGroundStateAdapter
from gpaw.response.site_data import AtomicSites, AtomicSiteData
from gpaw.response.localft import add_spin_polarization


@pytest.mark.response
def test_Fe_site_magnetization(gpw_files):
    # Set up ground state adapter
    calc = GPAW(gpw_files['fe_pw'], parallel=dict(domain=1))
    gs = ResponseGroundStateAdapter(calc)

    # Extract valid site radii range
    rmin_a, rmax_a = AtomicSiteData.valid_site_radii_range(gs)
    rmin = rmin_a[0]  # Only one magnetic atom in the unit cell
    rmax = rmax_a[0]
    # We expect rmax to be equal to the nearest neighbour distance
    # subtracted with the augmentation sphere radius. For a bcc lattice,
    # nn_dist = sqrt(3) a / 2:
    augr = gs.get_aug_radii()[0]
    rmax_expected = np.sqrt(3) * 2.867 / 2. - augr * Bohr
    assert abs(rmax - rmax_expected) < 1e-6
    # Test that an error is raised outside the valid range
    with pytest.raises(AssertionError):
        AtomicSiteData(
            gs, AtomicSites(indices=[0],  # Too small radii
                            radii=[np.linspace(rmin * 0.8, rmin, 5)]))
    with pytest.raises(AssertionError):
        AtomicSiteData(
            gs, AtomicSites(indices=[0],  # Too large radii
                            radii=[np.linspace(rmax, rmax * 1.2, 5)]))
    # Define atomic sites to span the valid range
    rc_r = np.linspace(rmin_a[0], rmax_a[0], 100)
    # Add the radius of the augmentation sphere explicitly
    rc_r = np.append(rc_r, [augr * Bohr])
    sites = AtomicSites(indices=[0], radii=[rc_r])
    site_data = AtomicSiteData(gs, sites)

    # Calculate site magnetization
    magmom_ar = site_data.calculate_magnetic_moments()
    magmom_r = magmom_ar[0]

    # Test that a cutoff at the augmentation sphere radius reproduces
    # the local magnetic moment of the GPAW calculation
    magmom_at_augr = calc.get_atoms().get_magnetic_moments()[0]
    assert abs(magmom_r[-1] - magmom_at_augr) < 4e-2

    # Do a manual calculation of the magnetic moment using the
    # all-electron partial waves
    # Calculate all-electron m(r)
    micro_setup = site_data.micro_setup_a[0]
    m_ng = np.array([micro_setup.rgd.zeros()
                     for n in range(micro_setup.Y_nL.shape[0])])
    for n, Y_L in enumerate(micro_setup.Y_nL):
        n_sg = np.dot(Y_L, micro_setup.n_sLg)
        add_spin_polarization(micro_setup.rgd, n_sg, m_ng[n, :])
    # Integrate with varrying radii
    m_g = integrate_lebedev(m_ng)
    ae_magmom_r = np.array([
        micro_setup.rgd.integrate_trapz(m_g, rcut=rcut / Bohr)
        for rcut in rc_r])
    # Test that values match approximately inside the augmentation sphere
    inaug_r = rc_r <= augr * Bohr
    assert magmom_r[inaug_r] == pytest.approx(ae_magmom_r[inaug_r], abs=3e-2)

    # import matplotlib.pyplot as plt
    # plt.plot(rc_r[:-1], magmom_r[:-1])
    # plt.plot(rc_r[:-1], ae_magmom_r[:-1], zorder=0)
    # plt.axvline(augr * Bohr, c='0.5', linestyle='--')
    # plt.xlabel(r'$r_\mathrm{c}$ [$\mathrm{\AA}$]')
    # plt.ylabel(r'$m$ [$\mu_\mathrm{B}$]')
    # plt.show()


@pytest.mark.response
def test_Co_site_data(gpw_files):
    # Set up ground state adapter
    calc = GPAW(gpw_files['co_pw'], parallel=dict(domain=1))
    gs = ResponseGroundStateAdapter(calc)

    # Extract valid site radii range
    rmin_a, rmax_a = AtomicSiteData.valid_site_radii_range(gs)
    # The valid ranges should be equal due to symmetry
    assert abs(rmin_a[1] - rmin_a[0]) < 1e-8
    assert abs(rmax_a[1] - rmax_a[0]) < 1e-8
    rmin = rmin_a[0]
    rmax = rmax_a[0]
    # We expect rmax to be equal to the nearest neighbour distance
    # subtracted with the augmentation sphere radius. For the hcp-lattice,
    # nn_dist = min(a, sqrt(a^2/3 + c^2/4)):
    augr_a = gs.get_aug_radii()
    assert abs(augr_a[1] - augr_a[0]) < 1e-8
    augr = augr_a[0]
    rmax_expected = min(2.5071, np.sqrt(2.5071**2 / 3 + 4.0695**2 / 4))
    rmax_expected -= augr * Bohr
    assert abs(rmax - rmax_expected) < 1e-6

    # Use radii spanning the entire valid range
    rc_r = np.linspace(rmin, rmax, 101)
    # Add the radius of the augmentation sphere explicitly
    rc_r = np.append(rc_r, [augr * Bohr])
    nr = len(rc_r)
    # Varry the site radii together and independently
    rc1_r = list(rc_r) + list(rc_r) + [augr * Bohr] * nr
    rc2_r = list(rc_r) + [augr * Bohr] * nr + list(rc_r)
    sites = AtomicSites(indices=[0, 1], radii=[rc1_r, rc2_r])
    site_data = AtomicSiteData(gs, sites)

    # Calculate site magnetization
    magmom_ar = site_data.calculate_magnetic_moments()

    # Test that the magnetization inside the augmentation sphere matches
    # the local magnetic moment of the GPAW calculation
    magmom_at_augr_a = calc.get_atoms().get_magnetic_moments()
    assert magmom_ar[:, -1] == pytest.approx(magmom_at_augr_a, abs=2e-2)

    # Test consistency of varrying radii
    assert magmom_ar[0, :nr] == pytest.approx(magmom_ar[1, :nr])
    assert magmom_ar[0, nr:2 * nr] == pytest.approx(magmom_ar[0, :nr])
    assert magmom_ar[0, 2 * nr:] == pytest.approx([magmom_ar[0, -1]] * nr)
    assert magmom_ar[1, nr:2 * nr] == pytest.approx([magmom_ar[1, -1]] * nr)
    assert magmom_ar[1, 2 * nr:] == pytest.approx(magmom_ar[1, :nr])

    # Calculate the atomic Zeeman energy
    rc_r = rc_r[:-1]
    sites = AtomicSites(indices=[0, 1], radii=[rc_r, rc_r])
    site_data = AtomicSiteData(gs, sites)
    EZ_ar = site_data.calculate_zeeman_energies()
    print(EZ_ar[0, ::20])

    # Test that the Zeeman energy comes out as expected
    assert EZ_ar[0] == pytest.approx(EZ_ar[1])
    assert EZ_ar[0, ::20] * 2 * Ha == pytest.approx([0.02638351, 1.41476112,
                                                     2.49540004, 2.79727200,
                                                     2.82727948, 2.83670767],
                                                    rel=1e-3)

    # import matplotlib.pyplot as plt
    # plt.subplot(1, 2, 1)
    # plt.plot(rc_r, magmom_ar[0, :nr - 1])
    # plt.axvline(augr * Bohr, c='0.5', linestyle='--')
    # plt.xlabel(r'$r_\mathrm{c}$ [$\mathrm{\AA}$]')
    # plt.ylabel(r'$m$ [$\mu_\mathrm{B}$]')
    # plt.subplot(1, 2, 2)
    # plt.plot(rc_r, EZ_ar[0] * Ha)  # Hartree -> eV
    # plt.axvline(augr * Bohr, c='0.5', linestyle='--')
    # plt.xlabel(r'$r_\mathrm{c}$ [$\mathrm{\AA}$]')
    # plt.ylabel(r'$E_\mathrm{Z}$ [eV]')
    # plt.show()



import pytest
import numpy as np

from ase import Atoms
from gpaw import GPAW, PW
from gpaw.mpi import world
from gpaw.test import findpeak
from gpaw.utilities import compiled_with_sl
from gpaw.response.df import DielectricFunction

# Comparing the plasmon peaks found in bulk sodium for two different
# atomic structures. Testing for identical plasmon peaks. Not using
# physical sodium cell.


@pytest.mark.dielectricfunction
@pytest.mark.response
def test_response_na_plasmon(in_tmp_dir):
    a = 4.23 / 2.0
    a1 = Atoms('Na',
               scaled_positions=[[0, 0.1, 0]],
               cell=(a, a, a),
               pbc=True)

    # parallel calculations must have domain = 1
    parallel = {'band': 1}
    if world.size > 1 and compiled_with_sl():
        parallel.update({'domain': 1})

    # Expanding along x-direction
    a2 = Atoms('Na2',
               scaled_positions=[[0, 0.1, 0], [0.5, 0.1, 0]],
               cell=(2 * a, a, a),
               pbc=True)

    a1.calc = GPAW(mode=PW(250),
                   kpts={'size': (4, 4, 4), 'gamma': True},
                   parallel=parallel,
                   # txt='small.txt',
                   )

    # Kpoint sampling should be halved in the expanded direction.
    a2.calc = GPAW(mode=PW(250),
                   kpts={'size': (2, 4, 4), 'gamma': True},
                   parallel=parallel,
                   # txt='large.txt',
                   )

    a1.get_potential_energy()
    a2.get_potential_energy()

    # Use twice as many bands for expanded structure
    a1.calc.diagonalize_full_hamiltonian(nbands=20)
    a2.calc.diagonalize_full_hamiltonian(nbands=40)

    a1.calc.write('gs_Na_small.gpw', 'all')
    a2.calc.write('gs_Na_large.gpw', 'all')

    # Settings that should yield the same result
    settings = [{'disable_point_group': True, 'disable_time_reversal': True},
                {'disable_point_group': False, 'disable_time_reversal': True},
                {'disable_point_group': True, 'disable_time_reversal': False},
                {'disable_point_group': False, 'disable_time_reversal': False}]

    # Test block parallelization (needs scalapack)
    if world.size > 1 and compiled_with_sl():
        settings.append({'disable_point_group': False,
                         'disable_time_reversal': False,
                         'nblocks': 2})

    # Calculate the dielectric functions
    dfs0 = []  # Arrays to check for self-consistency
    dfs1 = []
    dfs2 = []
    dfs3 = []
    dfs4 = []
    dfs5 = []

    # list of intensities to compare against. Intensity values matched
    # to 10-3 w/ higher tol. Speeding up test degraded the agreement to 10-2
    # Added additional intensity difference check test with tol 10-3
    I_diffs = {'x': [0.008999, 0.007558, 0.008999, 0.007558, 0.006101],
               'y': [0.004063, 0.004063, 0.004063, 0.004063, 0.004063],
               'z': [0.004063, 0.005689, 0.004244, 0.005689, 0.005689]}
    for idx, kwargs in enumerate(settings):
        df1 = DielectricFunction('gs_Na_small.gpw',
                                 ecut=40,
                                 rate=0.001,
                                 **kwargs)

        df1NLFCx, df1LFCx = df1.get_dielectric_function(direction='x')
        df1NLFCy, df1LFCy = df1.get_dielectric_function(direction='y')
        df1NLFCz, df1LFCz = df1.get_dielectric_function(direction='z')

        df2 = DielectricFunction('gs_Na_large.gpw',
                                 ecut=40,
                                 rate=0.001,
                                 **kwargs)

        df2NLFCx, df2LFCx = df2.get_dielectric_function(direction='x')
        df2NLFCy, df2LFCy = df2.get_dielectric_function(direction='y')
        df2NLFCz, df2LFCz = df2.get_dielectric_function(direction='z')

        dfs0.append(df1NLFCx)
        dfs1.append(df1LFCx)
        dfs2.append(df1NLFCy)
        dfs3.append(df1LFCy)
        dfs4.append(df1NLFCz)
        dfs5.append(df1LFCz)

        # Compare plasmon frequencies and intensities: x, y, z
        # x values
        w_w = df1.chi0calc.wd.omega_w
        w1, I1 = findpeak(w_w, -(1. / df1LFCx).imag)
        w2, I2 = findpeak(w_w, -(1. / df2LFCx).imag)
        I_diff = abs(I1 - I2)
        # test that the frequency for 2 settings are aprx equal
        assert w1 == pytest.approx(w2, abs=1e-2)
        # test that the intensity difference is within some tol
        assert I_diff == pytest.approx(I_diffs['x'][idx], abs=5e-3)
        # test that the intensities are aprx equal
        assert I1 == pytest.approx(I2, abs=1e-2)

        # y values
        w1, I1 = findpeak(w_w, -(1. / df1LFCy).imag)
        w2, I2 = findpeak(w_w, -(1. / df2LFCy).imag)
        I_diff = abs(I1 - I2)
        assert w1 == pytest.approx(w2, abs=1e-2)
        assert I_diff == pytest.approx(I_diffs['y'][idx], abs=5e-3)
        assert I1 == pytest.approx(I2, abs=1e-2)

        # z values
        w1, I1 = findpeak(w_w, -(1. / df1LFCz).imag)
        w2, I2 = findpeak(w_w, -(1. / df2LFCz).imag)
        I_diff = abs(I1 - I2)
        assert w1 == pytest.approx(w2, abs=1e-2)
        assert I_diff == pytest.approx(I_diffs['z'][idx], abs=5e-3)
        assert I1 == pytest.approx(I2, abs=1e-2)

    # Check for self-consistency
    for i, dfs in enumerate([dfs0, dfs1, dfs2, dfs3, dfs4, dfs5]):
        while len(dfs):
            df = dfs.pop()
            for df2 in dfs:
                try:
                    assert np.max(np.abs((df - df2) / df)) < 2e-3
                except AssertionError:
                    print(np.max(np.abs((df - df2) / df)))
                    raise AssertionError



"""Extensions to the ase Atoms class

"""
from ase import Atoms
from ase.build.connected import connected_indices
from ase.utils import deprecated

from .utilities.adjust_cell import adjust_cell


class Cluster(Atoms):
    """A class for cluster structures
    to enable simplified manipulation"""

    def __init__(self, *args, **kwargs):

        self.data = {}

        if len(args) > 0:
            filename = args[0]
            if isinstance(filename, str):
                self.read(filename, kwargs.get('filetype'))
                return
        else:
            Atoms.__init__(self, [])

        if kwargs.get('filename') is not None:
            filename = kwargs.pop('filename')
            Atoms.__init__(self, *args, **kwargs)
            self.read(filename, kwargs.get('filetype'))
        else:
            Atoms.__init__(self, *args, **kwargs)

    @deprecated(
        'Please use connected_indices from ase.build.connected instead.')
    def find_connected(self, index, dmax=None, scale=1.5):
        """Find atoms connected to self[index] and return them."""
        return self[connected_indices(self, index, dmax, scale)]

    @deprecated(
        'Please use adjust_cell from gpaw.utilities.adjust_cell instead.')
    def minimal_box(self, border=4, h=None, multiple=4) -> None:
        adjust_cell(self, border, h, multiple)



import pytest
import numpy as np

from ase.units import Hartree
from gpaw import GPAW
from gpaw.lcaotddft import LCAOTDDFT
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter
from gpaw.tddft.spectrum import photoabsorption_spectrum as spec_td
from gpaw.lrtddft import LrTDDFT
from gpaw.lrtddft import photoabsorption_spectrum as spec_lr
from gpaw.lrtddft2 import LrTDDFT2
from gpaw.mpi import world


pytestmark = [pytest.mark.usefixtures('module_tmp_path')]


@pytest.fixture(scope='module')
def time_propagation_calculation(gpw_files):
    td_calc = LCAOTDDFT(gpw_files['na2_tddft_sz'],
                        txt='na2_tddft_sz_td.out')
    DipoleMomentWriter(td_calc, 'dm.dat')
    td_calc.absorption_kick([0, 0, 1e-5])
    td_calc.propagate(30, 150)
    spec_td('dm.dat', 'spec_td.dat',
            e_min=0, e_max=10, width=0.5, delta_e=0.1)
    world.barrier()

    # Scale energy out due to \omega vs \omega_I difference in
    # broadened spectra in RT-TDDFT and LR-TDDFT
    data_ej = np.loadtxt('spec_td.dat')
    spec_e = data_ej[:, 3]
    spec_e[1:] /= data_ej[1:, 0]
    return spec_e


@pytest.fixture(scope='module')
def lrtddft_calculation(gpw_files):
    calc = GPAW(gpw_files['na2_tddft_sz'], txt=None)
    lr = LrTDDFT(calc, xc='LDA', txt='lr.out')
    lr.diagonalize()
    spec_lr(lr, 'spec_lr.dat',
            e_min=0, e_max=10, width=0.5, delta_e=0.1)
    world.barrier()

    # Scale energy out due to \omega vs \omega_I difference in
    # broadened spectra in RT-TDDFT and LR-TDDFT
    data_ej = np.loadtxt('spec_lr.dat')
    spec_e = data_ej[:, 4]
    spec_e[1:] /= lr[0].get_energy() * Hartree
    return spec_e


@pytest.fixture(scope='module')
def lrtddft2_calculation(gpw_files):
    calc = GPAW(gpw_files['na2_tddft_sz'], txt='lr2.out')
    lr2 = LrTDDFT2('lr2', calc, fxc='LDA')
    lr2.calculate()
    lr2.get_spectrum('spec_lr2.dat', 0, 10.1, 0.1, width=0.5)
    world.barrier()

    # Scale energy out due to \omega vs \omega_I difference in
    # broadened spectra in RT-TDDFT and LR-TDDFT
    data_ej = np.loadtxt('spec_lr2.dat')
    spec_e = data_ej[:, 5]
    spec_e[1:] /= lr2.lr_transitions.get_transitions()[0][0]
    return spec_e


@pytest.mark.rttddft
def test_lcaotddft_vs_lrtddft(time_propagation_calculation,
                              lrtddft_calculation):
    # One can decrease the tolerance by decreasing the time step
    # and other parameters
    assert (time_propagation_calculation
            == pytest.approx(lrtddft_calculation, abs=1e-2))


@pytest.mark.rttddft
def test_lcaotddft_vs_lrtddft2(time_propagation_calculation,
                               lrtddft2_calculation):
    # One can decrease the tolerance by decreasing the time step
    # and other parameters
    assert (time_propagation_calculation
            == pytest.approx(lrtddft2_calculation, abs=1e-2))


@pytest.mark.rttddft
def test_lrtddft_vs_lrtddft2(lrtddft_calculation,
                             lrtddft2_calculation):
    assert (lrtddft_calculation
            == pytest.approx(lrtddft2_calculation, abs=1e-3))



import pytest
from gpaw.mpi import serial_comm
from gpaw import GPAW
from gpaw.xc.rpa import RPACorrelation


@pytest.mark.rpa
@pytest.mark.response
@pytest.mark.parametrize('gpwfilename, eref', [
    ('si_qpoint_rounding_bug', -18.814),
    ('si_pw', -12.61),
])
def test_rpa_energy_si(in_tmp_dir, gpw_files, gpwfilename, eref):
    calc = GPAW(gpw_files[gpwfilename], communicator=serial_comm)
    calc.diagonalize_full_hamiltonian(nbands=50)

    ecut = 50
    rpa = RPACorrelation(calc, qsym=False, nfrequencies=8, ecut=[ecut])
    E_rpa_noqsym = rpa.calculate()

    rpa = RPACorrelation(calc, qsym=True, nfrequencies=8, ecut=[ecut])
    E_rpa_qsym = rpa.calculate()

    assert E_rpa_qsym == pytest.approx(eref, abs=0.01)
    assert E_rpa_qsym == pytest.approx(E_rpa_noqsym)



import numpy as np
import pytest

from gpaw.lcaotddft import LCAOTDDFT
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter
from gpaw.lcaotddft.wfwriter import WaveFunctionWriter
from gpaw.mpi import world
from gpaw.tddft.spectrum import photoabsorption_spectrum


@pytest.mark.rttddft
def test_lcaotddft_replay(gpw_files, in_tmp_dir):
    # Time-propagation calculation
    td_calc = LCAOTDDFT(gpw_files['na2_tddft_dzp'], txt='td.out')
    DipoleMomentWriter(td_calc, 'dm.dat')
    WaveFunctionWriter(td_calc, 'wf.ulm')
    WaveFunctionWriter(td_calc, 'wf_split.ulm', split=True)
    td_calc.absorption_kick(np.ones(3) * 1e-5)
    td_calc.propagate(20, 3)
    td_calc.write('td.gpw', mode='all')
    td_calc.propagate(7, 3)

    # Restart from the restart point
    td_calc = LCAOTDDFT('td.gpw', txt='td2.out')
    DipoleMomentWriter(td_calc, 'dm.dat')
    WaveFunctionWriter(td_calc, 'wf.ulm')
    WaveFunctionWriter(td_calc, 'wf_split.ulm')
    td_calc.propagate(20, 3)
    td_calc.propagate(20, 3)
    td_calc.propagate(10, 3)
    photoabsorption_spectrum('dm.dat', 'spec.dat')

    world.barrier()
    ref_i = np.loadtxt('spec.dat').ravel()

    # Replay both wf*.ulm files
    for tag in ['', '_split']:
        td_calc = LCAOTDDFT(gpw_files['na2_tddft_dzp'], txt='rep%s.out' % tag)
        DipoleMomentWriter(td_calc, 'dm_rep%s.dat' % tag)
        td_calc.replay(name='wf%s.ulm' % tag, update='density')
        photoabsorption_spectrum('dm_rep%s.dat' % tag, 'spec_rep%s.dat' % tag)

        world.barrier()

        # Check the spectrum files
        # Do this instead of dipolemoment files in order to see that the kick
        # was also written correctly in replaying
        data_i = np.loadtxt('spec_rep%s.dat' % tag).ravel()

        tol = 1e-10
        assert data_i == pytest.approx(ref_i, abs=tol)



from __future__ import annotations
import numpy as np
from numpy.linalg import eigvals
from typing import List, Tuple
from gpaw.typing import Array1D, Array2D


null_pole_thr = 1e-5
pole_resolution = 1e-5
epsilon = 1e-8

#  -------------------------------------------------------------
#  Old reference code
#  -------------------------------------------------------------

# 1 pole


def Xeval(Omega_GGp, residues_GGp, omega_w):
    X_GGpw = (
        residues_GGp[..., :, np.newaxis] * 2 * Omega_GGp[..., :, np.newaxis] /
        (omega_w[None, None, None, :]**2 - Omega_GGp[..., :, np.newaxis]**2)
    )

    return np.sum(X_GGpw, axis=2)


def mpa_cond1(z: tuple[complex, complex] | Array1D,
              E2: tuple[complex] | Array1D) -> \
        tuple[[complex], [float]] | Array2D:
    PPcond_rate = 0
    if abs(E2) < null_pole_thr:  # need to check also NAN(abs(E))
        PPcond_rate = 1
    elif np.real(E2) > 0.:
        pass
    else:
        PPcond_rate = 1

    # DALV: since MPA uses complex poles we need to guarantee the time ordering
    E2 = complex(abs(E2.real), -abs(E2.imag))
    E = np.emath.sqrt(E2)

    return E, PPcond_rate


def mpa_E_1p_solver(z, x):
    E2 = (x[0] * z[0]**2 - x[1] * z[1]**2) / (x[0] - x[1])
    E, PPcond_rate = mpa_cond1(z, E2)
    return E, PPcond_rate


def pole_is_out(i, wmax, thr, E):
    is_out = False

    if np.real(E[i]) > wmax:
        is_out = True

    j = 0
    while j < i and not is_out:
        if abs(np.real(E[i]) - np.real(E[j])) < thr:
            is_out = True
            if abs(np.real(E[j])) > max(abs(np.imag(E[j])),
               abs(np.imag(E[i]))):
                E[j] = np.mean([np.real(E[j]), np.real(E[i])]) - 1j * \
                    max(abs(np.imag(E[j])), abs(np.imag(E[i])))
            else:
                E[j] = np.mean([np.real(E[j]), np.real(E[i])]) - 1j * \
                    min(abs(np.imag(E[j])), abs(np.imag(E[i])))
        j = j + 1

    return is_out


def mpa_cond(npols: int, z: List[complex], E) ->\
        Tuple[int, List[bool], List[complex]]:
    PPcond = np.full(npols, False)
    npr = npols
    wmax = np.max(np.real(np.emath.sqrt(z))) * 1.5
    Eaux = np.emath.sqrt(E)

    i = 0
    while i < npr:
        Eaux[i] = max(abs(np.real(Eaux[i])), abs(np.imag(Eaux[i]))) - 1j * \
            min(abs(np.real(Eaux[i])), abs(np.imag(Eaux[i])))
        is_out = pole_is_out(i, wmax, pole_resolution, Eaux)

        if is_out:
            Eaux[i] = np.emath.sqrt(E[npr - 1])
            Eaux[i] = max(abs(np.real(Eaux[i])), abs(np.imag(Eaux[i]))) - 1j \
                * min(abs(np.real(Eaux[i])), abs(np.imag(Eaux[i])))
            PPcond[npr - 1] = True
            npr = npr - 1
        else:
            i = i + 1

    E[:npr] = Eaux[:npr]
    if npr < npols:
        E[npr:npols] = 2 * wmax - 0.01j
        PPcond[npr:npols] = True

    return E, npr, PPcond


def mpa_R_1p_fit(npols, npr, w, x, E):
    # Transforming the problem into a 2* larger least square with real numbers:
    A = np.zeros((4, 2), dtype='complex64')
    b = np.zeros((4), dtype='complex64')
    for k in range(2):
        b[2 * k] = np.real(x[k])
        b[2 * k + 1] = np.imag(x[k])
        A[2 * k][0] = 2. * np.real(E / (w[k]**2 - E**2))
        A[2 * k][1] = -2. * np.imag(E / (w[k]**2 - E**2))
        A[2 * k + 1][0] = 2. * np.imag(E / (w[k]**2 - E**2))
        A[2 * k + 1][1] = 2. * np.real(E / (w[k]**2 - E**2))

    Rri = np.linalg.lstsq(A, b, rcond=None)[0]
    R = Rri[0] + 1j * Rri[1]
    return R


def mpa_R_fit(npols, npr, w, x, E):
    # Transforming the problem into a 2* larger least square with real numbers:
    A = np.zeros((2 * npols * 2, npr * 2), dtype='complex64')
    b = np.zeros((2 * npols * 2), dtype='complex64')
    for k in range(2 * npols):
        b[2 * k] = np.real(x[k])
        b[2 * k + 1] = np.imag(x[k])
        for i in range(npr):
            A[2 * k][2 * i] = 2. * np.real(E[i] / (w[k]**2 - E[i]**2))
            A[2 * k][2 * i + 1] = -2. * np.imag(E[i] / (w[k]**2 - E[i]**2))
            A[2 * k + 1][2 * i] = 2. * np.imag(E[i] / (w[k]**2 - E[i]**2))
            A[2 * k + 1][2 * i + 1] = 2. * np.real(E[i] / (w[k]**2 - E[i]**2))

    Rri = np.linalg.lstsq(A, b, rcond=None)[0]
    R = np.zeros(npols, dtype='complex64')
    R[:npr] = Rri[::2] + 1j * Rri[1::2]
    return R


def mpa_E_solver_Pade(npols, z, x):
    b_m1 = b = np.zeros(npols + 1, dtype='complex64')
    b_m1[0] = b[0] = 1
    c = np.copy(x)

    for i in range(1, 2 * npols):

        c_m1 = np.copy(c)
        c[i:] = (c_m1[i - 1] - c_m1[i:]) / ((z[i:] - z[i - 1]) * c_m1[i:])

        b_m2 = np.copy(b_m1)
        b_m1 = np.copy(b)

        b = b_m1 - z[i - 1] * c[i] * b_m2
        b_m2[npols:0:-1] = c[i] * b_m2[npols - 1::-1]
        b[1:] = b[1:] + b_m2[1:]

    Companion = np.polynomial.polynomial.polycompanion(b[:npols + 1])
    # DALV: /b[npols] it is carried inside

    E = eigvals(Companion)

    # DALV: here we need to force real(E) to be positive.
    # This is because of the way the residue integral is performed, later.
    E, npr, PPcond = mpa_cond(npols, z, E)

    return E, npr, PPcond


def mpa_RE_solver(npols, w, x):
    if npols == 1:  # DALV: we could particularize the solution for 2-3 poles
        E, PPcond_rate = mpa_E_1p_solver(w, x)
        R = mpa_R_1p_fit(1, 1, w, x, E)
        # DALV: if PPcond_rate=0, R = x[1]*(z[1]**2-E**2)/(2*E)
        MPred = 0
    else:
        # DALV: Pade-Thiele solver (mpa_sol='PT')
        E, npr, PPcond = mpa_E_solver_Pade(npols, w**2, x)
        R = mpa_R_fit(npols, npr, w, x, E)

        PPcond_rate = 1
        MPred = 1

    # for later: MP_err = err_func_X(np, R, E, w, x)

    return R, E, MPred, PPcond_rate  # , MP_err



import numpy as np
from ase.units import Bohr

from gpaw.fd_operators import Laplace, Gradient
from gpaw.kpoint import KPoint
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.lfc import LocalizedFunctionsCollection as LFC
from gpaw.mpi import serial_comm
from gpaw.preconditioner import Preconditioner
from gpaw.projections import Projections
from gpaw.transformers import Transformer
from gpaw.utilities.blas import axpy
from gpaw.wavefunctions.arrays import UniformGridWaveFunctions
from gpaw.wavefunctions.fdpw import FDPWWaveFunctions
from gpaw.wavefunctions.mode import Mode
import gpaw.cgpaw as cgpaw


class FD(Mode):
    name = 'fd'

    def __init__(self, nn=3, interpolation=3, force_complex_dtype=False):
        self.nn = nn
        self.interpolation = interpolation
        Mode.__init__(self, force_complex_dtype)

    def __call__(self, *args, **kwargs):
        return FDWaveFunctions(self.nn, *args, **kwargs)

    def todict(self):
        dct = Mode.todict(self)
        dct['nn'] = self.nn
        dct['interpolation'] = self.interpolation
        return dct


class FDWaveFunctions(FDPWWaveFunctions):
    mode = 'fd'

    def __init__(self, stencil, parallel, initksl,
                 gd, nvalence, setups, bd,
                 dtype, world, kd, kptband_comm, timer, reuse_wfs_method=None,
                 collinear=True):
        FDPWWaveFunctions.__init__(self, parallel, initksl,
                                   reuse_wfs_method=reuse_wfs_method,
                                   collinear=collinear,
                                   gd=gd, nvalence=nvalence, setups=setups,
                                   bd=bd, dtype=dtype, world=world, kd=kd,
                                   kptband_comm=kptband_comm, timer=timer)

        # Kinetic energy operator:
        self.kin = Laplace(self.gd, -0.5, stencil, self.dtype)

        self.taugrad_v = None  # initialized by MGGA functional
        self.read_from_file_init_wfs_dm = False

    def empty(self, n=(), global_array=False, realspace=False, q=-1):
        return self.gd.empty(n, self.dtype, global_array)

    def integrate(self, a_xg, b_yg=None, global_integral=True):
        return self.gd.integrate(a_xg, b_yg, global_integral)

    def bytes_per_wave_function(self):
        return self.gd.bytecount(self.dtype)

    def set_setups(self, setups):
        self.pt = LFC(self.gd, [setup.pt_j for setup in setups],
                      self.kd, dtype=self.dtype, forces=True)
        FDPWWaveFunctions.set_setups(self, setups)

    def set_positions(self, spos_ac, atom_partition=None):
        FDPWWaveFunctions.set_positions(self, spos_ac, atom_partition)

    def __str__(self):
        s = 'Wave functions: Uniform real-space grid\n'
        s += '  Kinetic energy operator: %s\n' % self.kin.description
        return s + FDPWWaveFunctions.__str__(self)

    def make_preconditioner(self, block=1):
        return Preconditioner(self.gd, self.kin, self.dtype, block)

    def apply_pseudo_hamiltonian(self, kpt, ham, psit_xG, Htpsit_xG):
        self.timer.start('Apply hamiltonian')
        self.kin.apply(psit_xG, Htpsit_xG, kpt.phase_cd)
        ham.apply_local_potential(psit_xG, Htpsit_xG, kpt.s)
        ham.xc.apply_orbital_dependent_hamiltonian(
            kpt, psit_xG, Htpsit_xG, ham.dH_asp)
        self.timer.stop('Apply hamiltonian')

    def get_pseudo_partial_waves(self):
        phit_aj = [setup.get_partial_waves_for_atomic_orbitals()
                   for setup in self.setups]
        return LFC(self.gd, phit_aj, kd=self.kd, cut=True, dtype=self.dtype)

    def add_to_density_from_k_point_with_occupation(self, nt_sG, kpt, f_n):
        # Used in calculation of response part of GLLB-potential
        nt_G = nt_sG[kpt.s]
        for f, psit_G in zip(f_n, kpt.psit_nG):
            # Same as nt_G += f * abs(psit_G)**2, but much faster:
            cgpaw.add_to_density(f, psit_G, nt_G)

        # Hack used in delta-scf calculations:
        if hasattr(kpt, 'c_on'):
            assert self.bd.comm.size == 1
            d_nn = np.zeros((self.bd.mynbands, self.bd.mynbands),
                            dtype=complex)
            for ne, c_n in zip(kpt.ne_o, kpt.c_on):
                d_nn += ne * np.outer(c_n.conj(), c_n)
            for d_n, psi0_G in zip(d_nn, kpt.psit_nG):
                for d, psi_G in zip(d_n, kpt.psit_nG):
                    if abs(d) > 1.e-12:
                        nt_G += (psi0_G.conj() * d * psi_G).real

    def calculate_kinetic_energy_density(self):
        if self.taugrad_v is None:
            self.taugrad_v = [
                Gradient(self.gd, v, n=3, dtype=self.dtype).apply
                for v in range(3)]

        assert not hasattr(self.kpt_u[0], 'c_on')
        if not isinstance(self.kpt_u[0].psit_nG, np.ndarray):
            return None

        taut_sG = self.gd.zeros(self.nspins)
        dpsit_G = self.gd.empty(dtype=self.dtype)
        for kpt in self.kpt_u:
            for f, psit_G in zip(kpt.f_n, kpt.psit_nG):
                for v in range(3):
                    self.taugrad_v[v](psit_G, dpsit_G, kpt.phase_cd)
                    axpy(0.5 * f, abs(dpsit_G)**2, taut_sG[kpt.s])

        self.kptband_comm.sum(taut_sG)
        for taut_G in taut_sG:
            self.kd.symmetry.symmetrize(taut_G, self.gd)
        return taut_sG

    def apply_mgga_orbital_dependent_hamiltonian(self, kpt, psit_xG,
                                                 Htpsit_xG, dH_asp,
                                                 dedtaut_G):
        a_G = self.gd.empty(dtype=psit_xG.dtype)
        for psit_G, Htpsit_G in zip(psit_xG, Htpsit_xG):
            for v in range(3):
                self.taugrad_v[v](psit_G, a_G, kpt.phase_cd)
                self.taugrad_v[v](dedtaut_G * a_G, a_G, kpt.phase_cd)
                axpy(-0.5, a_G, Htpsit_G)

    def ibz2bz(self, atoms):
        """Transform wave functions in IBZ to the full BZ."""

        assert self.kd.comm.size == 1

        # New k-point descriptor for full BZ:
        kd = KPointDescriptor(self.kd.bzk_kc, nspins=self.nspins)
        kd.set_communicator(serial_comm)

        self.pt = LFC(self.gd, [setup.pt_j for setup in self.setups],
                      kd, dtype=self.dtype)
        self.pt.set_positions(atoms.get_scaled_positions())

        self.initialize_wave_functions_from_restart_file()

        weight = 2.0 / kd.nspins / kd.nbzkpts

        # Build new list of k-points:
        kpt_qs = []
        kpt_u = []
        for k in range(kd.nbzkpts):
            kpt_s = []
            for s in range(self.nspins):
                # Index of symmetry related point in the IBZ
                ik = self.kd.bz2ibz_k[k]
                r, q = self.kd.get_rank_and_index(ik)
                assert r == 0
                kpt = self.kpt_qs[q][s]

                phase_cd = np.exp(2j * np.pi * self.gd.sdisp_cd *
                                  kd.bzk_kc[k, :, np.newaxis])

                # New k-point:
                kpt2 = KPoint(1.0 / kd.nbzkpts, weight, s, k, k, phase_cd)
                kpt2.f_n = kpt.f_n / kpt.weight / kd.nbzkpts * 2 / self.nspins
                kpt2.eps_n = kpt.eps_n.copy()

                # Transform wave functions using symmetry operation:
                Psit_nG = self.gd.collect(kpt.psit_nG)
                if Psit_nG is not None:
                    Psit_nG = Psit_nG.copy()
                    for Psit_G in Psit_nG:
                        Psit_G[:] = self.kd.transform_wave_function(Psit_G, k)
                kpt2.psit = UniformGridWaveFunctions(
                    self.bd.nbands, self.gd, self.dtype,
                    kpt=k, dist=(self.bd.comm, self.bd.comm.size),
                    spin=kpt.s, collinear=True)
                self.gd.distribute(Psit_nG, kpt2.psit_nG)
                # Calculate PAW projections:
                nproj_a = [setup.ni for setup in self.setups]
                kpt2.projections = Projections(
                    self.bd.nbands, nproj_a,
                    kpt.projections.atom_partition,
                    self.bd.comm,
                    collinear=True, spin=s, dtype=self.dtype)

                kpt2.psit.matrix_elements(self.pt, out=kpt2.projections)
                kpt_s.append(kpt2)
                kpt_u.append(kpt2)
            kpt_qs.append(kpt_s)

        self.kd = kd
        self.kpt_qs = kpt_qs
        self.kpt_u = kpt_u

    def _get_wave_function_array(self, u, n, realspace=True, periodic=False):
        assert realspace
        kpt = self.kpt_u[u]
        psit_G = kpt.psit_nG[n]
        if periodic and self.dtype == complex:
            k_c = self.kd.ibzk_kc[kpt.k]
            return self.gd.plane_wave(-k_c) * psit_G
        return psit_G

    def write(self, writer, write_wave_functions=False):
        FDPWWaveFunctions.write(self, writer)

        if not write_wave_functions:
            return

        writer.add_array(
            'values',
            (self.nspins, self.kd.nibzkpts, self.bd.nbands) +
            tuple(self.gd.get_size_of_global_array()),
            self.dtype)

        for s in range(self.nspins):
            for k in range(self.kd.nibzkpts):
                for n in range(self.bd.nbands):
                    psit_G = self.get_wave_function_array(n, k, s)
                    writer.fill(psit_G * Bohr**-1.5)

    def read(self, reader):
        FDPWWaveFunctions.read(self, reader)

        if 'values' in reader.wave_functions:
            name = 'values'
        elif 'coefficients' in reader.wave_functions:
            name = 'coefficients'
        else:
            return

        c = reader.bohr**1.5
        if reader.version < 0:
            c = 1  # old gpw file

        for kpt in self.kpt_u:
            # We may not be able to keep all the wave
            # functions in memory - so psit_nG will be a special type of
            # array that is really just a reference to a file:
            psit_nG = reader.wave_functions.proxy(name, kpt.s, kpt.k)
            psit_nG.scale = c

            kpt.psit = UniformGridWaveFunctions(
                self.bd.nbands, self.gd, self.dtype, psit_nG,
                kpt=kpt.q, dist=(self.bd.comm, self.bd.comm.size),
                spin=kpt.s, collinear=True)

        if self.world.size > 1:
            # Read to memory:
            for kpt in self.kpt_u:
                kpt.psit.read_from_file()
                self.read_from_file_init_wfs_dm = True

    def initialize_from_lcao_coefficients(self, basis_functions):
        for kpt in self.kpt_u:
            kpt.psit = UniformGridWaveFunctions(
                self.bd.nbands, self.gd, self.dtype, kpt=kpt.q,
                dist=(self.bd.comm, self.bd.comm.size, 1),
                spin=kpt.s, collinear=True)
            kpt.psit_nG[:] = 0.0
            mynbands = len(kpt.C_nM)
            basis_functions.lcao_to_grid(kpt.C_nM,
                                         kpt.psit_nG[:mynbands], kpt.q)
            kpt.C_nM = None

    def random_wave_functions(self, nao):
        """Generate random wave functions."""

        gpts = self.gd.N_c[0] * self.gd.N_c[1] * self.gd.N_c[2]
        rng = np.random.default_rng(4 + self.world.rank)

        if self.bd.nbands < gpts / 64:
            gd1 = self.gd.coarsen()
            gd2 = gd1.coarsen()

            psit_G1 = gd1.empty(dtype=self.dtype)
            psit_G2 = gd2.empty(dtype=self.dtype)

            interpolate2 = Transformer(gd2, gd1, 1, self.dtype).apply
            interpolate1 = Transformer(gd1, self.gd, 1, self.dtype).apply

            shape = tuple(gd2.n_c)
            scale = np.sqrt(12 / gd2.volume)

            for kpt in self.kpt_u:
                for psit_G in kpt.psit_nG[nao:]:
                    if self.dtype == float:
                        psit_G2[:] = (rng.random(shape) - 0.5) * scale
                    else:
                        psit_G2.real = (rng.random(shape) - 0.5) * scale
                        psit_G2.imag = (rng.random(shape) - 0.5) * scale

                    interpolate2(psit_G2, psit_G1, kpt.phase_cd)
                    interpolate1(psit_G1, psit_G, kpt.phase_cd)

        elif gpts / 64 <= self.bd.nbands < gpts / 8:
            gd1 = self.gd.coarsen()

            psit_G1 = gd1.empty(dtype=self.dtype)

            interpolate1 = Transformer(gd1, self.gd, 1, self.dtype).apply

            shape = tuple(gd1.n_c)
            scale = np.sqrt(12 / gd1.volume)

            for kpt in self.kpt_u:
                for psit_G in kpt.psit_nG[nao:]:
                    if self.dtype == float:
                        psit_G1[:] = (rng.random(shape) - 0.5) * scale
                    else:
                        psit_G1.real = (rng.random(shape) - 0.5) * scale
                        psit_G1.imag = (rng.random(shape) - 0.5) * scale

                    interpolate1(psit_G1, psit_G, kpt.phase_cd)

        else:
            shape = tuple(self.gd.n_c)
            scale = np.sqrt(12 / self.gd.volume)

            for kpt in self.kpt_u:
                for psit_G in kpt.psit_nG[nao:]:
                    if self.dtype == float:
                        psit_G[:] = (rng.random(shape) - 0.5) * scale
                    else:
                        psit_G.real = (rng.random(shape) - 0.5) * scale
                        psit_G.imag = (rng.random(shape) - 0.5) * scale

    def estimate_memory(self, mem):
        FDPWWaveFunctions.estimate_memory(self, mem)



import pytest

import numpy as np

from gpaw import GPAW
from gpaw.mpi import world
from gpaw.response import ResponseContext, ResponseGroundStateAdapter
from gpaw.response.pw_parallelization import block_partition
from gpaw.response.matrix_elements import TransversePairPotentialCalculator

from gpaw.test.response.test_parallel_kptpair_extraction import \
    initialize_extractor, initialize_transitions, initialize_integral


@pytest.mark.response
@pytest.mark.later
def test_nicl2_pair_potential(gpw_files):
    """Test that the transverse pair potential vanishes in vacuum."""

    # ---------- Inputs ---------- #

    wfs = 'nicl2_pw_evac'
    q_qc = [[0., 0., 0.],
            [1. / 3., 1. / 3., 0.]]
    rshewmin = 1e-8
    nblocks = world.size // 2 if world.size % 2 == 0 else 1

    # ---------- Script ---------- #

    context = ResponseContext()
    calc = GPAW(gpw_files[wfs], parallel=dict(domain=1))
    gs = ResponseGroundStateAdapter(calc)

    # Set up extractor and transitions
    tcomm, kcomm = block_partition(context.comm, nblocks)
    extractor = initialize_extractor(gs, context, tcomm, kcomm)
    transitions = initialize_transitions(extractor, '+-', nbands=20)

    # Set up calculator
    pair_potential_calc = TransversePairPotentialCalculator(gs, context,
                                                            rshewmin=rshewmin)

    # Loop over k-point pairs
    for q_c in q_qc:
        integral = initialize_integral(extractor, context, q_c)
        for kptpair, weight in integral.weighted_kpoint_pairs(transitions):
            if weight is None:
                assert kptpair is None
                continue
            _, _, ut1_mytR, ut2_mytR = \
                pair_potential_calc.extract_pseudo_waves(kptpair)
            wt_mytR = \
                pair_potential_calc._evaluate_pseudo_matrix_element(
                    ut1_mytR, ut2_mytR)

            # Average out the in-plane degrees of freedom
            wt_mytz = np.average(wt_mytR, axis=(1, 2))

            # Gather all the transitions
            wt_tz = kptpair.tblocks.all_gather(wt_mytz)

            # import matplotlib.pyplot as plt
            # for wt_z in wt_tz:
            #     plt.subplot(1, 2, 1)
            #     plt.plot(np.arange(len(wt_z)), wt_z.real)
            #     plt.subplot(1, 2, 2)
            #     plt.plot(np.arange(len(wt_z)), wt_z.imag)
            # plt.show()

            # Find the maximum absolute value of the pair potential and make
            # sure that the pair potential is much smaller than that close to
            # the cell boundary for all the transitions
            abswt_tz = np.abs(wt_tz)
            max_wt = np.max(abswt_tz)
            assert np.all(abswt_tz[:, :10] < max_wt / 25.)
            assert np.all(abswt_tz[:, -10:] < max_wt / 25.)



from __future__ import annotations

import numpy as np
from scipy.special import spherical_jn
from dataclasses import dataclass

from gpaw.spline import Spline
from gpaw.ffbt import rescaled_fourier_bessel_transform
from gpaw.gaunt import gaunt, super_gaunt
from gpaw.spherical_harmonics import Y
from gpaw.atom.radialgd import RadialGridDescriptor
from gpaw.sphere.rshe import RealSphericalHarmonicsExpansion
from gpaw.response.pw_parallelization import Blocks1D


# Important note: The test suite monkeypatches this value to 2**10 so
# you may get different results in tests and production until we
# implement a better solution (e.g. generating setup-dependent radial_points).
#
# The motivation for lowering to 2**10 in tests is that many tests
# take 3-4 times longer if we do not.
#
# See https://gitlab.com/gpaw/gpaw/-/issues/984
DEFAULT_RADIAL_POINTS = 2**12


@dataclass
class LeanPAWDataset:
    rgd: RadialGridDescriptor
    l_j: np.ndarray
    rcut_j: np.ndarray
    phit_jg: np.ndarray
    phi_jg: np.ndarray
    # Number of radial points in spline interpolation
    radial_points: int | None = None

    def __post_init__(self):
        if self.radial_points is None:
            # We assign this late due to monkeypatch in testing
            self.radial_points = DEFAULT_RADIAL_POINTS

        # Number of basis functions
        self.ni = np.sum([2 * l + 1 for l in self.l_j])
        # Maximum angular momentum index l
        self.lmax = np.max(self.l_j)
        # Grid cutoff to create spline representation
        self.gcut2 = self.rgd.ceil(2 * max(self.rcut_j))

        # Set up cache
        self.dn_g_cache = {}
        self.dn_kspline_cache = {}

    def dn_kspline(self, j1: int, j2: int, l: int):
        """Get spline representation of Δn_jj'l(k)."""
        if (j1, j2, l) not in self.dn_kspline_cache:
            dn_g = self.get_pair_density_correction(j1, j2)
            self.dn_kspline_cache[j1, j2, l] = \
                self.rescaled_fourier_bessel_transform(dn_g, l)
        return self.dn_kspline_cache[j1, j2, l]

    def get_pair_density_correction(self, j1: int, j2: int):
        """Get Δn_jj'(r), while keeping the newest correction cached."""
        if (j1, j2) not in self.dn_g_cache:
            self.dn_g_cache = {}  # keep only one density in cache
            self.dn_g_cache[j1, j2] = self.calculate_pair_density_correction(
                j1, j2)
        return self.dn_g_cache[j1, j2]

    def calculate_pair_density_correction(self, j1, j2):
        """Calculate the pair density PAW correction for two partial waves.
                               ˷     ˷
        Δn (r) = φ (r) φ (r) - φ (r) φ (r)
          jj'     j     j'      j     j'
        """
        # (Real) radial functions for the partial waves
        phi_jg = self.phi_jg
        phit_jg = self.phit_jg
        return phi_jg[j1] * phi_jg[j2] - phit_jg[j1] * phit_jg[j2]

    def rescaled_fourier_bessel_transform(self, f_g, l):
        """Calculate the rescaled Fourier Bessel transform f_l(k)

                    rc
                4π  ⌠  2
        f (k) = ‾‾‾ ⎪ r dr j (kr) f(r)
         l      k^l ⌡       l
                    0
        """
        # First, we make a spline representation of the radial function f(r),
        # which is rescaled by a factor of r^-l by the radial grid descriptor:
        spline = self.rgd.spline(
            f_g[:self.gcut2], l=l, points=self.radial_points)
        # Once rescaled by r^-l, we can use the gpaw.ffbt module to Fast
        # Fourier Bessel Transform the radial spline r^-l f(r) and obtain
        # f_l(k) in a reciprocal spline representation
        kspline = rescaled_fourier_bessel_transform(
            spline, N=4 * self.radial_points)
        # Since this procedures hinges on a series of hardcoded parameters, we
        # return a self-testing version of the spline. If someone wants to run
        # calculations without dynamic testing of the FFBT methodology (i.e.
        # in an "unprotected" mode), simply return the bare `kspline` here.
        return SelfTestingKSpline(self.rgd, f_g, kspline)


class SelfTestingKSpline(Spline):
    """Self-testing reciprocal spline representation, f_l(k)."""

    def __init__(self, rgd, f_g, spline: Spline):
        # Store original real-space representation of the radial function f(r)
        self.rgd = rgd
        self.f_g = f_g
        super().__init__(spline.spline)

    def map(self, k_G):
        self.test_spline_representation(k_G)
        return super().map(k_G)

    def test_spline_representation(self, k_G):
        """Test validity of the FFBT implementation on input domain.

        At present, LeanPAWDataset's FFBT implementation relies on a range of
        hardcoded parameters, which are not guaranteed to work for all cases.

        In particular, the uniform radial grid used for the FFBT is defined
        through the `rcut` and `N` parameters in
        `rescaled_fourier_bessel_transform()`
        where the former is hardcoded inside the function itself.

        Furthermore, the `points` parameter to `rgd.spline()` controls the
        fidelity of the interpolation between nonlinear and equidistant radial
        grids needed to make use of the FFBT algorithm.

        To make a generally reliable implementation, one would need to control
        all of these parameters based on the setup, e.g. the nonlinear radial
        grid spacing. In doing so, one should be mindful that the `rcut`
        parameter defines the reciprocal grid spacing of the kspline and that
        `N` controls the range of the reciprocal space domain.

        For now, we simply check that the requested plane waves are within the
        computed k-range of the FFBT and check that the resulting transforms
        match a manual calculation at a few selected K-vectors.
        """
        kmax = np.max(k_G)
        assert kmax <= self.get_cutoff()

        # Manual calculation at finite k
        k_k = np.array([kmax, np.average(k_G)])
        f_k = 4 * np.pi * fourier_bessel_transform(
            np.array(k_k), self.l, self.rgd, self.f_g)
        # Manual calculation at k=0
        if self.l == 0:  # Vanishes for l>0
            k_k = np.append(k_k, [0.])
            f_k = np.append(f_k, [self.rgd.integrate(self.f_g)])
        # FFBT calculation
        myf_k = k_k**self.l * super().map(k_k)
        assert np.allclose(myf_k, f_k, rtol=1e-2, atol=1e-3), \
            f'FFBT mismatch: {myf_k}, {f_k}'


def calculate_pair_density_correction(qG_Gv: np.ndarray, *,
                                      pawdata: LeanPAWDataset):
    r"""Calculate the atom-centered PAW correction to the pair density.
                                                      ˍ
    The atom-centered pair density correction tensor, Q_aii', is defined as the
    atom-centered Fourier transform

    ˍ             /                                     ˷         ˷
    Q_aii'(G+q) = | dr e^-i(G+q)r [φ_ai^*(r) φ_ai'(r) - φ_ai^*(r) φ_ai'(r)]
                  /

    evaluated with the augmentation sphere center at the origin. The full pair
    density correction tensor is then given by
                                  ˍ
    Q_aii'(G+q) = e^(-i[G+q].R_a) Q_aii'(G+q)

    Expanding the plane wave coefficient into real spherical harmonics and
    spherical Bessel functions, the correction can split into angular and
    radial contributions

                    l
                __  __
                \   \      l  m ˰   m,m_i,m_i'
    Q_aii'(K) = /   /  (-i)  Y (K) g
                ‾‾  ‾‾        l     l,l_i,l_i'
                l  m=-l
                            rc
                            /                    a     a      ˷a    ˷a
                       × 4π | r^2 dr j_l(|K|r) [φ (r) φ (r) - φ (r) φ (r)]
                            /                    j_i   j_i'    j_i   j_i'
                            0

    where K = G+q and g denotes the Gaunt coefficients.

    For more information, see [PRB 103, 245110 (2021)]. In particular, it
    should be noted that the partial waves themselves are defined via real
    spherical harmonics and radial functions φ_j from the PAW setups:

     a       m_i ˰   a
    φ (r) = Y   (r) φ (r)
     i       l_i     j_i
    """
    ni = pawdata.ni  # Number of partial waves
    l_j = pawdata.l_j  # l-index for each radial function index j
    G_LLL = gaunt(pawdata.lmax)

    # Initialize correction tensor
    npw = qG_Gv.shape[0]
    Qbar_Gii = np.zeros((npw, ni, ni), dtype=complex)

    # K-vector norm
    k_G = np.linalg.norm(qG_Gv, axis=1)

    # Loop of radial function indices for partial waves i and i'
    i1_counter = 0
    for j1, l1 in enumerate(l_j):
        i2_counter = 0
        for j2, l2 in enumerate(l_j):
            # Sample l according to the Gaunt coefficient selection rules, see
            # e.g. gpaw.test.test_gaunt
            for l in range(abs(l1 - l2), l1 + l2 + 1, 2):
                # Calculate the spherical Fourier-Bessel transform
                #                  rc
                #              4π  /
                # Δn_jj'l(k) = ‾‾‾ | r^2 dr j_l(kr) Δn_jj'(r)
                #              k^l /
                #                  0
                # To evaluate the radial integral efficiently, we rely on the
                # Fast Fourier Bessel Transform (FFBT) algorithm, see gpaw.ffbt
                # The FFBT produces a radial spline representation of
                # Δn_jj'l(k), which we map to the K-vector norms in question:
                dn_G = pawdata.dn_kspline(j1, j2, l).map(k_G)

                # Angular part of the integral
                f_G = (-1j)**l * dn_G
                for m in range(l**2, (l + 1)**2):
                    # Calculate the solid harmonic
                    #        m ˰
                    # |K|^l Y (K)
                    #        l
                    klY_G = Y(m, *qG_Gv.T)

                    # Generate m-indices for each radial function
                    for m1 in range(2 * l1 + 1):
                        for m2 in range(2 * l2 + 1):
                            # Set up the i=(l,m) index for each partial wave
                            i1 = i1_counter + m1
                            i2 = i2_counter + m2
                            # Extract Gaunt coefficients
                            gaunt_coeff = G_LLL[l1**2 + m1, l2**2 + m2, m]
                            if (gaunt_coeff == 0):
                                continue

                            # Add contribution to the PAW correction
                            Qbar_Gii[:, i1, i2] += gaunt_coeff * klY_G * f_G

            # Add to i and i' counters
            i2_counter += 2 * l2 + 1
        i1_counter += 2 * l1 + 1
    return Qbar_Gii


def calculate_matrix_element_correction(qG_Gv, pawdata,
                                        rshe: RealSphericalHarmonicsExpansion):
    r"""Calculate the atom-centered correction to a generalized matrix element.

    For matrix elements corresponding to the expectation value of a plane wave
    coefficient e^-i(G+q)r and a known functional of the (spin-)density
    f[n](r), the PAW correction tensor is given by

    F_aii'(G+q) = <φ_ai| e^-i(G+q)r f[n](r) |φ_ai'>
                     ˷                         ˷
                  - <φ_ai| e^-i(G+q)r f[n](r) |φ_ai'>
                                  ˍ
                = e^(-i[G+q].R_a) F_aii'(G+q)
          ˍ
    where F_aii'(G+q) is the atom-centered PAW correction tensor.

    Expanding the functional f[n](r) in the atom-centered frame in real
    spherical harmonics (corresponding to the input rshe),

                  l
              __  __
         →    \   \   m ˰   m
    f[n](r) = /   /  Y (r) f (r)
              ‾‾  ‾‾  l     l
              l  m=-l

    expansion of the plane-wave coefficient in real spherical harmonics and
    spherical Bessel functions j_l(Kr) yields the following expression for the
    atom-centered correction tensor [publication in preparation]:

                       l        l'
                   __  __   __  __
    ˍ              \   \    \   \      l'  m'˰   m_i,m_i',m,m'
    F_aii'(K) = 4π /   /    /   /  (-i)   Y (K) G
                   ‾‾  ‾‾   ‾‾  ‾‾         l'    l_i,l_i',l,l'
                   l  m=-l  l' m'=-l'

                                rc
                                /  2            a     a      ˷a    ˷a      m
                              × | r dr j (Kr) [φ (r) φ (r) - φ (r) φ (r)] f (r)
                                /       l'      j_i   j_i'    j_i   j_i'   l
                                0

    where K=G+q and G_LLLL denotes the super Gaunt coefficients, which yield
    the integrals over four spherical harmonics.
    """
    rgd = rshe.rgd
    assert rgd is pawdata.xc_correction.rgd
    ni = pawdata.ni  # Number of partial waves
    l_j = pawdata.l_j  # l-index for each radial function index j
    lmax = max(l_j)
    assert max(rshe.l_M) <= 2 * lmax
    G_LLLL = super_gaunt(lmax)
    # (Real) radial functions for the partial waves
    phi_jg = pawdata.phi_jg
    phit_jg = pawdata.phit_jg
    # Truncate the radial functions to span only the radial grid coordinates
    # which need correction
    assert np.allclose(rgd.r_g, pawdata.rgd.r_g[:rgd.N])
    phi_jg = np.array(phi_jg)[:, :rgd.N]
    phit_jg = np.array(phit_jg)[:, :rgd.N]

    # Initialize correction tensor
    npw = qG_Gv.shape[0]
    Fbar_Gii = np.zeros((npw, ni, ni), dtype=complex)

    # K-vector norm and direction
    k_G = np.linalg.norm(qG_Gv, axis=1)
    Kd_Gv = qG_Gv.copy()
    Kd_Gv[k_G > 1e-10] /= k_G[k_G > 1e-10, np.newaxis]

    # Loop of radial function indices for partial waves i and i'
    i1_counter = 0
    for j1, l1 in enumerate(l_j):
        i2_counter = 0
        for j2, l2 in enumerate(l_j):
            # Calculate the radial partial wave correction
            #                              ˷      ˷
            # Δn_jj'(r) = φ_j(r) φ_j'(r) - φ_j(r) φ_j'(r)
            dn_g = phi_jg[j1] * phi_jg[j2] - phit_jg[j1] * phit_jg[j2]

            # Loop through the angular components in the real spherical
            # harmonics expansion of f[n](r)
            for l, L, f_g in zip(rshe.l_M, rshe.L_M, rshe.f_gM.T):
                dnf_g = dn_g * f_g
                # Apply Gaunt coefficient selection rules to loop through
                # the l' coefficients of the plane-wave expansion
                lpmin = np.min(abs(
                    np.arange(abs(l1 - l2), l1 + l2 + 1) - l))
                for lp in range(lpmin, l1 + l2 + l + 1):
                    if not (l1 + l2 + l + lp) % 2 == 0:
                        continue
                    # Calculate radial part of the correction
                    dnf_G = parallel_fourier_bessel_transform(
                        k_G, lp, rgd, dnf_g)

                    # Calculate angular part of the correction
                    x_G = 4 * np.pi * (-1j)**lp * dnf_G
                    # Loop through available m-indices for the partial waves
                    # and generate the composite L=(l,m) index as well as the
                    # partial wave index i
                    for m1 in range(2 * l1 + 1):
                        L1 = l1**2 + m1
                        i1 = i1_counter + m1
                        for m2 in range(2 * l2 + 1):
                            L2 = l2**2 + m2
                            i2 = i2_counter + m2
                            # Loop through m' indices of the plane-wave
                            # expansion and generate the L' composite index
                            for mp in range(2 * lp + 1):
                                Lp = lp**2 + mp
                                # If the angular integral (super gaunt
                                # coefficient) is finite,
                                coeff = G_LLLL[L1, L2, L, Lp]
                                if abs(coeff) > 1e-10:
                                    # Calculate spherical harmonic and add
                                    # contribution to the PAW correction
                                    Y_G = Y(Lp, *Kd_Gv.T)
                                    Fbar_Gii[:, i1, i2] += coeff * Y_G * x_G

            # Add to i and i' counters
            i2_counter += 2 * l2 + 1
        i1_counter += 2 * l1 + 1
    return Fbar_Gii


def parallel_fourier_bessel_transform(k_G, *args, comm=None):
    """Distribute FBT plane-wave components over a given communicator."""
    # NB: If we need to do something similar elsewhere, we can generalize this
    # function to a decorator!
    if comm is None:
        from gpaw.mpi import world as comm
    Gblocks = Blocks1D(comm, len(k_G))
    f_myG = fourier_bessel_transform(k_G[Gblocks.myslice], *args)
    return Gblocks.all_gather(f_myG)


def fourier_bessel_transform(k_G, l, rgd, f_g):
    """Perform a spherical Fourier-Bessel transform of a radial function f(r).

    Computes the transform

            max
           r
           ⌠  2
    f(k) = ⎪ r dr j (kr) f(r)
           ⌡       l
           0

    on the supplied radial grid.
    """
    # Vectorize calculation of spherical Bessel functions
    l_Gg = l * np.ones((len(k_G), rgd.N), dtype=int)
    kr_Gg = k_G[:, np.newaxis] * rgd.r_g[np.newaxis]
    jl_Gg = spherical_jn(l_Gg, kr_Gg)  # so slow...
    # Integrate the radial grid using linear interpolation
    f_G = rgd.integrate_trapz(jl_Gg * f_g[np.newaxis])
    return f_G


class PWPAWCorrectionData:
    def __init__(self, Q_aGii, qpd, pawdatasets, pos_av, atomrotations):
        # Sometimes we loop over these in ways that are very dangerous.
        # It must be list, not dictionary.
        assert isinstance(Q_aGii, list)
        assert len(Q_aGii) == len(pos_av) == len(pawdatasets.by_atom)

        self.Q_aGii = Q_aGii

        self.qpd = qpd
        self.pawdatasets = pawdatasets
        self.pos_av = pos_av
        self.atomrotations = atomrotations

    def _new(self, Q_aGii):
        return PWPAWCorrectionData(Q_aGii, qpd=self.qpd,
                                   pawdatasets=self.pawdatasets,
                                   pos_av=self.pos_av,
                                   atomrotations=self.atomrotations)

    def remap(self, M_vv, G_Gv, sym, sign):
        Q_aGii = []
        for a, Q_Gii in enumerate(self.Q_aGii):
            x_G = self._get_x_G(G_Gv, M_vv, self.pos_av[a])
            U_ii = self.atomrotations.get_by_a(a).R_sii[sym]

            Q_Gii = np.einsum('ij,kjl,ml->kim',
                              U_ii,
                              Q_Gii * x_G[:, None, None],
                              U_ii,
                              optimize='optimal')
            if sign == -1:
                Q_Gii = Q_Gii.conj()
            Q_aGii.append(Q_Gii)

        return self._new(Q_aGii)

    def _get_x_G(self, G_Gv, M_vv, pos_v):
        # This doesn't really belong here.  Or does it?  Maybe this formula
        # is only used with PAW corrections.
        return np.exp(1j * (G_Gv @ (pos_v - M_vv @ pos_v)))

    def remap_by_symop(self, symop, G_Gv, M_vv):
        return self.remap(M_vv, G_Gv, symop.symno, symop.sign)

    def multiply(self, P_ani, band):
        assert isinstance(P_ani, list)
        assert len(P_ani) == len(self.Q_aGii)

        C1_aGi = [Qa_Gii @ P1_ni[band].conj()
                  for Qa_Gii, P1_ni in zip(self.Q_aGii, P_ani)]
        return C1_aGi

    def reduce_ecut(self, G2G):
        # XXX actually we should return this with another PW descriptor.
        return self._new([Q_Gii.take(G2G, axis=0) for Q_Gii in self.Q_aGii])

    def almost_equal(self, otherpawcorr, G_G):
        for a, Q_Gii in enumerate(otherpawcorr.Q_aGii):
            e = abs(self.Q_aGii[a] - Q_Gii[G_G]).max()
            if e > 1e-12:
                return False
        return True


def get_pair_density_paw_corrections(pawdatasets, qpd, spos_ac, atomrotations):
    r"""Calculate and bundle paw corrections to the pair densities as a
    PWPAWCorrectionData object.

    The pair density PAW correction tensor is given by:

                  /
    Q_aii'(G+q) = | dr e^(-i[G+q].r) [φ_ai^*(r-R_a) φ_ai'(r-R_a)
                  /                     ˷             ˷
                                      - φ_ai^*(r-R_a) φ_ai'(r-R_a)]
    """
    qG_Gv = qpd.get_reciprocal_vectors(add_q=True)
    pos_av = spos_ac @ qpd.gd.cell_cv

    # Calculate pair density PAW correction tensor
    Qbar_xGii = {}
    for species_index, pawdata in pawdatasets.by_species.items():
        # Calculate atom-centered correction tensor
        Qbar_Gii = calculate_pair_density_correction(qG_Gv, pawdata=pawdata)
        # Add dependency on the atomic position (phase factor)
        Qbar_xGii[species_index] = Qbar_Gii

    Q_aGii = []
    for a, (pos_v, pawdata) in enumerate(zip(pos_av, pawdatasets.by_atom)):
        x_G = np.exp(-1j * (qG_Gv @ pos_v))
        species_index = pawdatasets.id_by_atom[a]
        Qbar_Gii = Qbar_xGii[species_index]
        Q_aGii.append(x_G[:, np.newaxis, np.newaxis] * Qbar_Gii)

    return PWPAWCorrectionData(Q_aGii, qpd=qpd,
                               pawdatasets=pawdatasets,
                               pos_av=pos_av,
                               atomrotations=atomrotations)


def get_matrix_element_paw_corrections(qpd, pawdata_a, rshe_a, spos_ac):
    r"""Calculate the PAW correction to a generalized matrix element.

    For a given functional of the electron (spin-)density f[n](r), the PAW
    correction is given by
                                  ˍ
    F_aii'(G+q) = e^(-i[G+q].R_a) F_aii'(G+q)
          ˍ
    where F_aii'(G+q) is the atom-centered correction (see above).
    """
    qG_Gv = qpd.get_reciprocal_vectors(add_q=True)

    F_aGii = []
    for pawdata, rshe, spos_c in zip(pawdata_a.by_atom, rshe_a, spos_ac):
        # Calculate atom-centered PAW correction
        Fbar_Gii = calculate_matrix_element_correction(
            qG_Gv, pawdata, rshe)

        # XXX Can time be saved by doing some of the processing per species
        # rather than per atom?

        # Add dependency on the atomic position (phase factor)
        pos_v = spos_c @ qpd.gd.cell_cv
        x_G = np.exp(-1j * (qG_Gv @ pos_v))
        F_aGii.append(x_G[:, np.newaxis, np.newaxis] * Fbar_Gii)

    return F_aGii



import numpy as np

from ase.units import Hartree, Bohr

from gpaw.directmin.tools import get_n_occ
from gpaw.utilities import unpack_hermitian
from gpaw.xc.hybrid import HybridXCBase


def calculate_forces(wfs, dens, ham, log=None):
    """Return the atomic forces."""

    assert not isinstance(ham.xc, HybridXCBase)
    assert not ham.xc.name.startswith('GLLB')

    func_name = None
    if hasattr(wfs.eigensolver, 'dm_helper'):
        func_name = getattr(wfs.eigensolver.dm_helper.func, 'name', None)
    elif hasattr(wfs.eigensolver, 'odd'):
        func_name = getattr(wfs.eigensolver.odd, 'name', None)
    if func_name == 'PZ-SIC':
        if wfs.mode == 'fd' or wfs.mode == 'pw':
            return calculate_forces_using_non_diag_lagr_matrix(
                wfs, dens, ham, log)
        elif wfs.mode == 'lcao':
            for kpt in wfs.kpt_u:
                kpt.rho_MM = \
                    wfs.calculate_density_matrix(kpt.f_n, kpt.C_nM)

    natoms = len(wfs.setups)

    # Force from projector functions (and basis set):
    F_wfs_av = np.zeros((natoms, 3))
    wfs.calculate_forces(ham, F_wfs_av)
    wfs.gd.comm.sum(F_wfs_av, 0)
    F_ham_av = np.zeros((natoms, 3))

    try:
        # ODD functionals need force corrections for each spin
        correction = ham.xc.setup_force_corrections
    except AttributeError:
        pass
    else:
        correction(F_ham_av)

    ham.calculate_forces(dens, F_ham_av)

    F_av = F_ham_av + F_wfs_av
    wfs.world.broadcast(F_av, 0)

    if func_name == 'PZ-SIC' and wfs.mode == 'lcao':
        F_av += wfs.eigensolver.dm_helper.func.get_odd_corrections_to_forces(
            wfs, dens)
        for kpt in wfs.kpt_u:
            # need to re-set rho_MM otherwise it will be used
            # it's probably better to in wfs.reset, but
            # when position changes wfs.reset is not called
            kpt.rho_MM = None

    F_av = wfs.kd.symmetry.symmetrize_forces(F_av)

    if log:
        log('\nForces in eV/Ang:')
        c = Hartree / Bohr
        for a, setup in enumerate(wfs.setups):
            log('%3d %-2s %10.5f %10.5f %10.5f' %
                ((a, setup.symbol) + tuple(F_av[a] * c)))
        log()

    return F_av


def calculate_forces_using_non_diag_lagr_matrix(wfs, dens, ham, log=None):

    natoms = len(wfs.setups)
    F_wfs_av = np.zeros((natoms, 3))
    esolv = wfs.eigensolver

    grad_knG = esolv.get_gradients_2(ham, wfs)
    if 'SIC' in esolv.odd.name:
        for kpt in wfs.kpt_u:
            esolv.odd.get_energy_and_gradients_kpt(
                wfs, kpt, grad_knG, esolv.iloop.U_k, add_grad=True)
    for kpt in wfs.kpt_u:
        k = esolv.n_kps * kpt.s + kpt.q
        n_occ = get_n_occ(kpt)[0]

        lamb = wfs.integrate(
            kpt.psit_nG[:n_occ], grad_knG[k][:n_occ], True)

        P_ani = kpt.P_ani
        dP_aniv = wfs.pt.dict(n_occ, derivative=True)
        wfs.pt.derivative(kpt.psit_nG[:n_occ], dP_aniv, kpt.q)
        dH_asp = ham.dH_asp

        for a, dP_niv in dP_aniv.items():
            dP_niv = dP_niv.conj()
            dO_ii = wfs.setups[a].dO_ii
            P_ni = P_ani[a][:n_occ]
            dS_nkv = np.einsum('niv,ij,kj->nkv', dP_niv, dO_ii, P_ni)
            dS_nkv = \
                (dS_nkv + np.transpose(dS_nkv, (1, 0, 2)).conj())
            # '-' becuase there is extra minus in 'pt derivative'
            F_wfs_av[a] -= np.einsum('kn,nkv->v', lamb, dS_nkv).real
            dH_ii = unpack_hermitian(dH_asp[a][kpt.s])
            dh_nv = np.einsum('niv,ij,nj->nv', dP_niv, dH_ii, P_ni)
            # '+' becuase there is extra minus in 'pt derivative'
            F_wfs_av[a] += np.einsum('n,nv->v', kpt.f_n[:n_occ],
                                     2.0 * dh_nv.real)

        if 'SIC' in esolv.odd.name:
            esolv.odd.get_odd_corrections_to_forces(F_wfs_av,
                                                    wfs,
                                                    kpt)

    wfs.kd.comm.sum(F_wfs_av, 0)
    wfs.gd.comm.sum(F_wfs_av, 0)

    F_ham_av = np.zeros((natoms, 3))
    ham.calculate_forces(dens, F_ham_av)
    F_av = F_ham_av + F_wfs_av
    wfs.world.broadcast(F_av, 0)

    F_av = wfs.kd.symmetry.symmetrize_forces(F_av)

    if log:
        log('\nForces in eV/Ang:')
        c = Hartree / Bohr
        for a, setup in enumerate(wfs.setups):
            log('%3d %-2s %10.5f %10.5f %10.5f' %
                ((a, setup.symbol) + tuple(F_av[a] * c)))
        log()

    return F_av



"""
tests solvation parameters as in

Damian A. Scherlis, Jean-Luc Fattebert, Francois Gygi,
Matteo Cococcioni and Nicola Marzari,
J. Chem. Phys. 124, 074103, 2006
"""

from gpaw import GPAW
from gpaw.utilities.adjust_cell import adjust_cell
import pytest
from ase import Atoms
from ase.units import mol, kcal, Pascal, m, Bohr
from gpaw.solvation import (
    SolvationGPAW,
    FG02SmoothStepCavity,
    LinearDielectric,
    GradientSurface,
    SurfaceInteraction,
    ElDensity
)


def test_solvation_sfgcm06():
    SKIP_VAC_CALC = True

    h = 0.24
    vac = 4.0

    epsinf = 78.36
    rho0 = 0.00078 / Bohr ** 3
    beta = 1.3
    st = 72. * 1e-3 * Pascal * m
    convergence = {
        'energy': 0.05 / 8.,
        'density': 10.,
        'eigenstates': 10.,
    }

    atoms = Atoms('Cl')
    adjust_cell(atoms, vac, h)

    if not SKIP_VAC_CALC:
        atoms.calc = GPAW(
            mode='fd', xc='PBE', h=h, charge=-1, convergence=convergence)
        Evac = atoms.get_potential_energy()
        print(Evac)
    else:
        # h=0.24, vac=4.0, setups: 0.9.11271, convergence: only energy 0.05 / 8
        Evac = -3.83245253419

    atoms.calc = SolvationGPAW(
        mode='fd', xc='PBE', h=h, charge=-1, convergence=convergence,
        cavity=FG02SmoothStepCavity(
            rho0=rho0, beta=beta,
            density=ElDensity(),
            surface_calculator=GradientSurface()),
        dielectric=LinearDielectric(epsinf=epsinf),
        interactions=[SurfaceInteraction(surface_tension=st)])
    Ewater = atoms.get_potential_energy()
    assert atoms.calc.get_number_of_iterations() < 40
    atoms.get_forces()
    DGSol = (Ewater - Evac) / (kcal / mol)
    print('Delta Gsol: %s kcal / mol' % DGSol)

    assert DGSol == pytest.approx(-75., abs=10.)



import numpy as np
import pytest
from gpaw.core.matrix import Matrix
from gpaw.gpu import cupy as cp, as_np, as_xp
from gpaw.mpi import world
from gpaw.new.c import GPU_AWARE_MPI
from gpaw.gpu.mpi import CuPyMPI


@pytest.mark.gpu
@pytest.mark.serial
def test_zyrk():
    a = np.array([[1, 1 + 2j, 2], [1, 0.5j, -1 - 0.5j]])
    m = Matrix(2, 3, data=a)
    b = m.multiply(m, opb='C', beta=0.0, symmetric=True)
    b.tril2full()
    a = cp.asarray(a)
    m = Matrix(2, 3, data=a)
    b2 = m.multiply(m, opb='C', beta=0.0, symmetric=True)
    b2.tril2full()
    c = b2.to_cpu()
    assert (c.data == b.data).all()


@pytest.mark.gpu
@pytest.mark.serial
def test_eigh():
    H1 = Matrix(2, 2, data=np.array([[2, 42.1 + 42.1j], [0.1 - 0.1j, 3]]))
    S1 = Matrix(2, 2, data=np.array([[1, 42.1 + 42.2j], [0.1 - 0.2j, 0.9]]))
    H2 = Matrix(2, 2, data=cp.asarray(H1.data))
    S2 = Matrix(2, 2, data=cp.asarray(S1.data))

    E1 = H1.eigh(S1)

    S0 = S1.copy()
    S0.tril2full()

    E2 = H2.eigh(S2)
    assert as_np(E2) == pytest.approx(E1)

    C1 = H1.data
    C2 = H2.to_cpu().data

    # Check that eigenvectors are parallel:
    X = C1.conj() @ S0.data @ C2.T
    assert abs(X) == pytest.approx(np.eye(2))


def op(a: np.ndarray, o: str) -> np.ndarray:
    if o == 'N':
        return a
    if o == 'C':
        return a.T.conj()
    1 / 0


@pytest.mark.gpu
@pytest.mark.parametrize(
    'shape1, shape2, op1, op2, sym, same',
    [((5, 9), (5, 9), 'N', 'C', 1, 1),
     ((2, 3), (2, 3), 'N', 'C', 1, 0),
     ((5, 9), (5, 9), 'N', 'C', 0, 0),
     ((5, 9), (5, 9), 'C', 'N', 0, 0),
     ((5, 9), (9, 5), 'C', 'C', 0, 0),
     ((5, 5), (5, 9), 'N', 'N', 0, 0)])
@pytest.mark.parametrize('beta', [0.0, 1.0])
@pytest.mark.parametrize('dtype', [float, complex])
@pytest.mark.parametrize('xp', [np, cp])
def test_mul(shape1, shape2, op1, op2, beta, sym, same, dtype, xp, rng):
    if world.size > 1 and xp is cp:
        if op1 == 'C' or (op1 == 'N' and op2 == 'C' and sym and beta == 0.0):
            pytest.skip('Not implemented!')
    alpha = 1.234
    comm = world if GPU_AWARE_MPI else CuPyMPI(world)

    shape3 = (shape1[0] if op1 == 'N' else shape1[1],
              shape2[1] if op2 == 'N' else shape2[0])
    m1, m2, m3 = (Matrix(*shape, dtype=dtype, dist=(comm, 1, 1), xp=xp)
                  for shape in [shape1, shape2, shape3])

    if world.rank == 0:
        for m in [m1, m2, m3]:
            data = m.data.view(float)
            data[:] = as_xp(rng.random(data.shape), xp)
        if sym:
            m2.data[:] = m1.data
            m3.data += m3.data.T.conj()

        # Correct result:
        a1, a2, a3 = (as_np(m.data) for m in [m1, m2, m3])
        a3 = beta * a3 + alpha * op(a1, op1) @ op(a2, op2)

    M1, M2, M3 = (Matrix(*shape, dtype=dtype, dist=(comm, -1, 1), xp=xp)
                  for shape in [shape1, shape2, shape3])
    for m, M in zip([m1, m2, m3], [M1, M2, M3]):
        m.redist(M)

    if same:
        M2 = M1

    M1.multiply(M2, alpha=alpha, opa=op1, opb=op2, beta=beta,
                out=M3, symmetric=sym)

    m3 = M3.gather()
    if world.rank == 0:
        if sym:
            m3.tril2full()
        error = abs(a3 - as_np(m3.data)).max()
    else:
        error = 0.0
    error = world.sum_scalar(error)
    assert error < 1e-13


if __name__ == '__main__':
    test_mul((1, 1), (1, 19), 'N', 'N', 0.0, 0, 0,
             complex, cp, np.random.default_rng(42))



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

"""Finite difference operators.

This file defines a series of finite difference operators used in grid mode.
"""

from math import pi, factorial as fact

import numpy as np
from numpy.fft import fftn, ifftn

import gpaw.cgpaw as cgpaw
from gpaw import debug
from gpaw.gpu import cupy_is_fake


# Expansion coefficients for finite difference Laplacian.  The numbers are
# from J. R. Chelikowsky et al., Phys. Rev. B 50, 11355 (1994):
laplace = [[0],
           [-2, 1],
           [-5 / 2, 4 / 3, -1 / 12],
           [-49 / 18, 3 / 2, -3 / 20, 1 / 90],
           [-205 / 72, 8 / 5, -1 / 5, 8 / 315, -1 / 560],
           [-5269 / 1800, 5 / 3, -5 / 21, 5 / 126, -5 / 1008, 1 / 3150],
           [-5369 / 1800, 12 / 7, -15 / 56, 10 / 189, -1 / 112, 2 / 1925,
            -1 / 16632]]

derivatives = [[1 / 2],
               [2 / 3, -1 / 12],
               [3 / 4, -3 / 20, 1 / 60],
               [4 / 5, -1 / 5, 4 / 105, -1 / 280]]


class FDOperator:
    def __init__(self, coef_p, offset_pc, gd, dtype=float,
                 description=None, xp=np):
        """FDOperator(coefs, offsets, gd, dtype) -> FDOperator object.
        """

        # Is this a central finite-difference type of stencil?
        cfd = True
        for offset_c in offset_pc:
            if sum([offset != 0 for offset in offset_c]) >= 2:
                cfd = False

        mp = np.abs(offset_pc).max()  # padding

        # If stencil is strictly larger than any domain, some CPU will
        # need data from the second-nearest-neighbour domain.
        # We don't support that.  (Typically happens with 1D distributions)
        assert (mp <= gd.n_c).all(), 'Stencil longer than domain'
        n_c = gd.n_c
        M_c = n_c + 2 * mp
        stride_c = np.array([M_c[1] * M_c[2], M_c[2], 1])
        offset_p = np.dot(offset_pc, stride_c)
        coef_p = np.ascontiguousarray(coef_p, float)
        neighbor_cd = gd.neighbor_cd

        assert coef_p.ndim == 1
        assert coef_p.shape == offset_p.shape
        assert dtype in [float, complex]

        self.dtype = dtype
        self.shape = tuple(n_c)

        if gd.comm.size > 1:
            comm = gd.comm.get_c_object()
        else:
            comm = None

        assert neighbor_cd.flags.c_contiguous and offset_p.flags.c_contiguous
        self.mp = mp
        self.gd = gd
        self.npoints = len(coef_p)
        self.coef_p = coef_p
        self.offset_p = offset_p
        self.offset_pc = offset_pc
        self.comm = comm
        self.cfd = cfd

        self.xp = xp
        gpu = xp is not np and not cupy_is_fake
        self.operator = cgpaw.Operator(coef_p, offset_p, n_c, mp,
                                       neighbor_cd, dtype == float,
                                       comm, cfd, gpu)

        if description is None:
            description = '%d point finite-difference stencil' % self.npoints
        self.description = description

    def __str__(self):
        return '<' + self.description + '>'

    def __call__(self, in_xR, out_xR=None):
        """New UGArray interface."""
        if out_xR is None:
            out_xR = in_xR.new()
        if self.xp is np:
            self.operator.apply(in_xR.data, out_xR.data,
                                in_xR.desc.phase_factor_cd)
        elif cupy_is_fake:
            self.operator.apply(in_xR.data._data, out_xR.data._data,
                                in_xR.desc.phase_factor_cd)
        else:
            assert isinstance(in_xR.data, self.xp.ndarray)
            assert isinstance(out_xR.data, self.xp.ndarray)
            self.operator.apply_gpu(in_xR.data.data.ptr, out_xR.data.data.ptr,
                                    in_xR.data.shape, in_xR.data.dtype,
                                    in_xR.desc.phase_factor_cd)
        return out_xR

    def apply(self, in_xg, out_xg, phase_cd=None):
        """Old NumPy interface."""
        if self.xp is np:
            self.operator.apply(in_xg, out_xg, phase_cd)
        elif cupy_is_fake:
            self.operator.apply(in_xg._data, out_xg._data,
                                phase_cd)
        else:
            self.operator.apply_gpu(in_xg.data.ptr,
                                    out_xg.data.ptr,
                                    in_xg.shape, in_xg.dtype, phase_cd)

    def relax(self, relax_method, f_g, s_g, n, w=None):
        if self.xp is np:
            self.operator.relax(relax_method, f_g, s_g, n, w)
        elif cupy_is_fake:
            self.operator.relax(relax_method, f_g._data, s_g._data, n, w)
        else:
            self.operator.relax_gpu(relax_method,
                                    f_g.data.ptr,
                                    s_g.data.ptr,
                                    n, w)

    def get_diagonal_element(self):
        return self.operator.get_diagonal_element()

    def get_async_sizes(self):
        return self.operator.get_async_sizes()


if debug:
    _FDOperator = FDOperator

    class FDOperator(_FDOperator):  # type: ignore
        def apply(self, in_xg, out_xg, phase_cd=None):
            assert in_xg.shape == out_xg.shape
            assert in_xg.shape[-3:] == self.shape
            assert in_xg.flags.c_contiguous
            assert in_xg.dtype == self.dtype
            assert out_xg.flags.c_contiguous
            assert out_xg.dtype == self.dtype
            assert (self.dtype == float or
                    (phase_cd.dtype == complex and
                     phase_cd.shape == (3, 2)))
            _FDOperator.apply(self, in_xg, out_xg, phase_cd)

        def relax(self, relax_method, f_g, s_g, n, w=None):
            assert f_g.shape == self.shape
            assert s_g.shape == self.shape
            assert f_g.flags.c_contiguous
            assert f_g.dtype == float
            assert s_g.flags.c_contiguous
            assert s_g.dtype == float
            assert self.dtype == float
            _FDOperator.relax(self, relax_method, f_g, s_g, n, w)


def Laplace(gd, scale=1.0, n=1, dtype=float, xp=np):
    if n == 9:
        return FTLaplace(gd, scale, dtype)
    else:
        return GUCLaplace(gd, scale, n, dtype, xp=xp)


class GUCLaplace(FDOperator):
    def __init__(self, gd, scale=1.0, n=1, dtype=float, xp=np):
        """Laplacian for general non orthorhombic grid.

        gd: GridDescriptor
            Descriptor for grid.
        scale: float
            Scaling factor.  Use scale=-0.5 for a kinetic energy operator.
        n: int
            Range of stencil.  Stencil has O(h^(2n)) error.
        dtype: float or complex
            Data-type to work on.
        """

        # Order the 26 neighbor grid points after length
        # (reduced to 13 inequivalent):
        M_ic = np.indices((3, 3, 3)).reshape((3, -3)).T[-13:] - 1
        u_cv = gd.h_cv / (gd.h_cv**2).sum(1)[:, np.newaxis]**0.5
        u2_i = (np.dot(M_ic, u_cv)**2).sum(1)
        i_d = u2_i.argsort()

        # x^2, y^2, z^2, yz, xz, xy:
        m_mv = np.array([(2, 0, 0), (0, 2, 0), (0, 0, 2),
                         (0, 1, 1), (1, 0, 1), (1, 1, 0)])
        # Try 3, 4, 5 and 6 of the shortest directions:
        for D in range(3, 7):
            h_dv = np.dot(M_ic[i_d[:D]], gd.h_cv)
            A_md = (h_dv**m_mv[:, np.newaxis, :]).prod(2)
            # Find best stencil coefficients:
            a_d, residual, rank, s = np.linalg.lstsq(A_md, [1, 1, 1, 0, 0, 0],
                                                     rcond=-1)
            if residual.sum() < 1e-14:
                if rank != D:
                    raise ValueError(
                        'You have a weird unit cell!  '
                        'Try to use the maximally reduced Niggli cell.  '
                        'See the ase.build.niggli_reduce() function.')
                # D directions was OK
                break

        a_d *= scale
        offsets = [(0, 0, 0)]
        coefs = [laplace[n][0] * a_d.sum()]
        for d in range(D):
            M_c = M_ic[i_d[d]]
            offsets.extend(np.arange(1, n + 1)[:, np.newaxis] * M_c)
            coefs.extend(a_d[d] * np.array(laplace[n][1:]))
            offsets.extend(np.arange(-1, -n - 1, -1)[:, np.newaxis] * M_c)
            coefs.extend(a_d[d] * np.array(laplace[n][1:]))

        FDOperator.__init__(self, coefs, offsets, gd, dtype, xp=xp)

        self.description = (
            '%d*%d+1=%d point O(h^%d) finite-difference Laplacian' %
            ((self.npoints - 1) // n, n, self.npoints, 2 * n))


class Gradient(FDOperator):
    def __init__(self, gd, v, scale=1.0, n=1, dtype=float, xp=np):
        """Symmetric gradient for general non orthorhombic grid.

        gd: GridDescriptor
            Descriptor for grid.
        v: int
            Direction of gradient: 0, 1, or 2 for x, y and z.
        scale: float
            Scaling factor.
        n: int
            Range of stencil.  Stencil has O(h^(2n)) error.
        dtype: float or complex
            Data-type to work on.
        """

        from scipy.spatial import Voronoi

        # Find nearest neighbors.  If h is a vector pointing at a
        # neighbor grid-points then we don't also include -h in the list:
        M_ic = np.indices((3, 3, 3)).reshape((3, -1)).T - 1
        h_iv = M_ic.dot(gd.h_cv)
        voro = Voronoi(h_iv)
        i_d = []  # List[int]
        for i1, i2 in voro.ridge_points:
            if i1 == 13 and i2 > 13:
                i_d.append(i2)
            elif i2 == 13 and i1 > 13:
                i_d.append(i1)

        D = len(i_d)  # number of neighbors (3, 4, 5, 6 or 7)

        h_dv = h_iv[i_d]  # vectors pointing at neighbor grid-points

        # Find gradient along 3 directions (n_cv):
        invh_vc = np.linalg.inv(gd.h_cv)
        n_cv = (invh_vc / (invh_vc**2).sum(axis=0)**0.5).T

        coef_cd = np.zeros((3, D))  # unknown weights
        for c, n_v in enumerate(n_cv):
            # Point on the plane perpendicular to the directions get a
            # weight of zero
            ok_d = abs(h_dv.dot(n_v)) > 1e-10
            h_mv = h_dv[ok_d]

            # The theee equations: A_jm.dot(coef_m) = [1, 0, 0]
            A_jm = np.array([h_mv.dot(n_v),
                             h_mv.dot(gd.h_cv[c - 1]),
                             h_mv.dot(gd.h_cv[c - 2])])

            U_jm, S_m, V_mm = np.linalg.svd(A_jm, full_matrices=False)
            coef_m = V_mm.T.dot(np.diag(S_m**-1).dot(U_jm.T.dot([1, 0, 0])))
            coef_cd[c, ok_d] = coef_m

        # Now get the v-gradient:
        coef_d = np.linalg.inv(n_cv)[v].dot(coef_cd) * scale

        # Create stencil:
        offsets = []
        coefs = []
        stencil = np.array(derivatives[n - 1])
        for d, c in enumerate(coef_d):
            if abs(c) < 1e-10:
                continue
            M_c = M_ic[i_d[d]]
            offsets.extend(np.arange(1, n + 1)[:, np.newaxis] * M_c)
            coefs.extend(c * stencil)
            offsets.extend(np.arange(-1, -n - 1, -1)[:, np.newaxis] * M_c)
            coefs.extend(-c * stencil)

        FDOperator.__init__(self, coefs, offsets, gd, dtype, xp=xp)

        self.description = (
            'Finite-difference {}-derivative with O(h^{}) error ({} points)'
            .format('xyz'[v], 2 * n, self.npoints))


class LaplaceA(FDOperator):
    def __init__(self, gd, scale, dtype=float, xp=np):
        assert gd.orthogonal
        c = np.divide(-1 / 12, gd.h_cv.diagonal()**2) * scale  # Why divide?
        c0 = c[1] + c[2]
        c1 = c[0] + c[2]
        c2 = c[1] + c[0]
        a = -16.0 * np.sum(c)
        b = 10.0 * c + 0.125 * a
        FDOperator.__init__(self,
                            [a,
                             b[0], b[0],
                             b[1], b[1],
                             b[2], b[2],
                             c0, c0, c0, c0,
                             c1, c1, c1, c1,
                             c2, c2, c2, c2],
                            [(0, 0, 0),
                             (-1, 0, 0), (1, 0, 0),
                             (0, -1, 0), (0, 1, 0),
                             (0, 0, -1), (0, 0, 1),
                             (0, -1, -1), (0, -1, 1), (0, 1, -1), (0, 1, 1),
                             (-1, 0, -1), (-1, 0, 1), (1, 0, -1), (1, 0, 1),
                             (-1, -1, 0), (-1, 1, 0), (1, -1, 0), (1, 1, 0)],
                            gd, dtype,
                            'O(h^4) Mehrstellen Laplacian (A)',
                            xp=xp)


class LaplaceB(FDOperator):
    def __init__(self, gd, dtype=float, xp=np):
        a = 0.5
        b = 1.0 / 12.0
        FDOperator.__init__(self,
                            [a,
                             b, b, b, b, b, b],
                            [(0, 0, 0),
                             (-1, 0, 0), (1, 0, 0),
                             (0, -1, 0), (0, 1, 0),
                             (0, 0, -1), (0, 0, 1)],
                            gd, dtype,
                            'O(h^4) Mehrstellen Laplacian (B)',
                            xp=xp)


class FTLaplace:
    def __init__(self, gd, scale, dtype):
        assert gd.comm.size == 1 and gd.pbc_c.all()

        N_c1 = gd.N_c[:, np.newaxis]
        i_cq = np.indices(gd.N_c).reshape((3, -1))
        i_cq += N_c1 // 2
        i_cq %= N_c1
        i_cq -= N_c1 // 2
        B_vc = 2.0 * pi * gd.icell_cv.T
        k_vq = np.dot(B_vc, i_cq)
        k_vq *= k_vq
        self.k2_Q = k_vq.sum(axis=0).reshape(gd.N_c)
        self.k2_Q *= -scale
        self.d = 6.0 / gd.h_cv[0, 0]**2
        self.npoints = 1000

    def apply(self, in_xg, out_xg, phase_cd=None):
        if in_xg.ndim > 3:
            for in_g, out_g in zip(in_xg, out_xg):
                out_g[:] = ifftn(fftn(in_g) * self.k2_Q).real
        else:
            out_xg[:] = ifftn(fftn(in_xg) * self.k2_Q).real

    def get_diagonal_element(self):
        return self.d


class OldGradient(FDOperator):
    def __init__(self, gd, v, scale=1.0, n=1, dtype=float, xp=np):
        h = (gd.h_cv**2).sum(1)**0.5
        d = gd.xxxiucell_cv[:, v]
        A = np.zeros((2 * n + 1, 2 * n + 1))
        for i, io in enumerate(range(-n, n + 1)):
            for j in range(2 * n + 1):
                A[i, j] = io**j / float(fact(j))
        A[n, 0] = 1.0
        coefs = np.linalg.inv(A)[1]
        coefs = np.delete(coefs, len(coefs) // 2)
        offs = np.delete(np.arange(-n, n + 1), n)
        coef_p = []
        offset_pc = []
        for i in range(3):
            if abs(d[i]) > 1e-11:
                coef_p.extend(list(coefs * d[i] / h[i] * scale))
                offset = np.zeros((2 * n, 3), int)
                offset[:, i] = offs
                offset_pc.extend(offset)

        FDOperator.__init__(self, coef_p, offset_pc, gd, dtype,
                            'O(h^%d) %s-gradient stencil' % (2 * n, 'xyz'[v]),
                            xp=xp)



from ase.build import molecule
from ase.parallel import parprint
from gpaw import GPAW
from gpaw.utilities.adjust_cell import adjust_cell
import pytest


def test_fixocc():
    h = 0.4
    box = 2
    nbands = 2
    txt = '-'
    txt = None

    H2 = molecule('H2')
    adjust_cell(H2, box, h)
    convergence = {'energy': 0.01, 'eigenstates': 0.001, 'density': 0.01}

    base_kwargs = dict(
        mode='fd',
        h=h,
        nbands=nbands,
        convergence=convergence,
        txt=txt)

    if 1:
        # test ZeroKelvin vs FixedOccupations
        c = GPAW(**base_kwargs, occupations={'width': 0.0})
        H2.calc = c
        E_zk = H2.get_potential_energy()

        c = GPAW(**base_kwargs,
                 occupations=dict(name='fixed', numbers=[[1, 0]]))
        H2.calc = c
        E_fo = H2.get_potential_energy()
        parprint(E_zk, E_fo)
        assert E_zk == pytest.approx(E_fo, abs=1.e-10)

    if 1:
        # test spin-paired vs spin-polarized
        c = GPAW(**base_kwargs,
                 occupations={'name': 'fixed', 'numbers': [[0.5, 0.5]]})
        H2.calc = c
        E_ns = H2.get_potential_energy()
    if 1:
        c = GPAW(**base_kwargs,
                 spinpol=True,
                 occupations={'name': 'fixed', 'numbers': [[0.5, 0.5]] * 2})
        H2.calc = c
        E_sp = H2.get_potential_energy()
        parprint(E_ns, E_sp)
        assert E_ns == pytest.approx(E_sp, abs=1.e-6)



"""Check for change in total energy and lowest eigenvalue regarding to box."""
from ase import Atoms
from ase.parallel import parprint
from gpaw import GPAW, PoissonSolver
from gpaw.utilities.adjust_cell import adjust_cell
import pytest


def test_generic_asym_box(in_tmp_dir):
    h = 0.2
    s = Atoms('He')
    adjust_cell(s, 3, h=h)

    c = GPAW(mode='fd', charge=1, txt='He_plus.txt',
             poissonsolver=PoissonSolver(use_charge_center=True),
             convergence={     # run fast
                 'energy': 0.001,
                 'eigenstates': 1e-4,
                 'density': 1e-3})
    s.calc = c
    e_small = s.get_potential_energy()
    eps_small = c.get_eigenvalues()[0]

    cell = s.get_cell()
    cell[0] *= 2
    s.set_cell(cell)
    e_big = s.get_potential_energy()
    eps_big = c.get_eigenvalues()[0]

    parprint('Energies and Eigenvalues:')
    parprint('     Small Box    Wide Box')
    parprint('E:   {0:9.3f}     {1:9.3f}'.format(e_small, e_big))
    parprint('eps: {0:9.3f}     {1:9.3f}'.format(eps_small, eps_big))
    assert e_small == pytest.approx(e_big, abs=2.5e-4)
    assert eps_small == pytest.approx(eps_big, abs=6e-4)



"""This module defines different external potentials."""
import copy
import warnings
from typing import Callable, Dict, Optional

import numpy as np
from ase.units import Bohr, Ha

import gpaw.cgpaw as cgpaw
from gpaw.typing import Array3D

__all__ = ['ConstantPotential', 'ConstantElectricField', 'CDFTPotential',
           'PointChargePotential', 'StepPotentialz',
           'PotentialCollection']

known_potentials: Dict[str, Callable] = {}


def _register_known_potentials():
    from gpaw.bfield import BField
    known_potentials['CDFTPotential'] = lambda: None  # ???
    known_potentials['BField'] = BField
    for name in __all__:
        known_potentials[name] = globals()[name]


def create_external_potential(name, **kwargs):
    """Construct potential from dict."""
    if not known_potentials:
        _register_known_potentials()
    return known_potentials[name](**kwargs)


class ExternalPotential:
    vext_g: Optional[Array3D] = None
    vext_q: Optional[Array3D] = None

    def get_potential(self, gd):
        """Get the potential on a regular 3-d grid.

        Will only call calculate_potential() the first time."""

        if self.vext_g is None:
            self.calculate_potential(gd)
            self.vext_g.flags.writeable = False
        return self.vext_g

    def get_potentialq(self, gd, pd3):
        """Get the potential on a regular 3-d grid in real space.

        Will only call calculate_potential() the first time."""

        if self.vext_q is None:
            vext_g = self.get_potential(gd)
            self.vext_q = pd3.fft(vext_g)
            self.vext_q.flags.writeable = False

        return self.vext_q

    def calculate_potential(self, gd) -> None:
        raise NotImplementedError

    def get_name(self) -> str:
        return self.__class__.__name__

    def update_potential_pw(self, ham, dens) -> float:
        v_q = self.get_potentialq(ham.finegd, ham.pd3).copy()
        eext = ham.pd3.integrate(v_q, dens.rhot_q, global_integral=False)
        dens.map23.add_to1(ham.vt_Q, v_q)
        ham.vt_sG[:] = ham.pd2.ifft(ham.vt_Q)
        if not ham.collinear:
            ham.vt_xG[1:] = 0.0
        return eext

    def update_atomic_hamiltonians_pw(self, ham, W_aL, dens) -> None:
        vext_q = self.get_potentialq(ham.finegd, ham.pd3)
        dens.ghat.integrate(ham.vHt_q + vext_q, W_aL)

    def paw_correction(self, Delta_p, dH_sp) -> None:
        pass

    def derivative_pw(self, ham, ghat_aLv, dens) -> None:
        vext_q = self.get_potentialq(ham.finegd, ham.pd3)
        dens.ghat.derivative(ham.vHt_q + vext_q, ghat_aLv)


class NoExternalPotential(ExternalPotential):
    vext_g = np.zeros((0, 0, 0))

    def __str__(self):
        return 'NoExternalPotential'

    def update_potential_pw(self, ham, dens) -> float:
        ham.vt_sG[:] = ham.pd2.ifft(ham.vt_Q)
        if not ham.collinear:
            ham.vt_xG[1:] = 0.0
        return 0.0

    def update_atomic_hamiltonians_pw(self, ham, W_aL, dens):
        dens.ghat.integrate(ham.vHt_q, W_aL)

    def derivative_pw(self, ham, ghat_aLv, dens):
        dens.ghat.derivative(ham.vHt_q, ghat_aLv)


class ConstantPotential(ExternalPotential):
    """Constant potential for tests."""
    def __init__(self, constant=1.0):
        self.constant = constant / Ha
        self.name = 'ConstantPotential'

    def __str__(self):
        return f'Constant potential: {(self.constant * Ha):.3f} V'

    def calculate_potential(self, gd):
        self.vext_g = gd.zeros() + self.constant

    def todict(self):
        return {'name': self.name,
                'constant': self.constant * Ha}


class ConstantElectricField(ExternalPotential):
    def __init__(self, strength, direction=[0, 0, 1], tolerance=1e-7):
        """External constant electric field.

        strength: float
            Field strength in V/Ang.
        direction: vector
            Polarization direction.
        """
        self.strength = strength * Bohr / Ha
        self.direction_v = np.array(direction, dtype=float)
        self.direction_v /= np.linalg.norm(self.direction_v)
        self.field_v = self.strength * self.direction_v
        self.tolerance = tolerance
        self.name = 'ConstantElectricField'

    def __str__(self):
        return ('Constant electric field: '
                '({:.3f}, {:.3f}, {:.3f}) V/Ang'
                .format(*(self.field_v * Ha / Bohr)))

    def calculate_potential(self, gd):
        # Note that PW-mode is periodic in all directions!
        L_c = abs(gd.cell_cv @ self.direction_v)
        # eps = self.tolerance
        # assert (L_c < eps).sum() == 2

        center_v = 0.5 * gd.cell_cv.sum(0)
        r_gv = gd.get_grid_point_coordinates().transpose((1, 2, 3, 0))
        f_g = (r_gv - center_v) @ self.direction_v

        # Set potential to zero at boundary of box (important for PW-mode).
        # Say we have 8 grid points.  Instead of a potential like this:
        #
        # -4 -3 -2 -1  0  1  2  3
        #
        # we want:
        #
        #  0 -3 -2 -1  0  1  2  3

        L = L_c.sum()
        f_g[abs(abs(f_g) - L / 2) < 1e-5] = 0.0

        self.vext_g = f_g * self.strength

    def todict(self):
        return {'name': self.name,
                'strength': self.strength * Ha / Bohr,
                'direction': self.direction_v}


class ProductPotential(ExternalPotential):
    def __init__(self, ext_i):
        self.ext_i = ext_i

    def calculate_potential(self, gd):
        self.vext_g = self.ext_i[0].get_potential(gd).copy()
        for ext in self.ext_i[1:]:
            self.vext_g *= ext.get_potential(gd)

    def __str__(self):
        return '\n'.join(['Product of potentials:'] +
                         [ext.__str__() for ext in self.ext_i])

    def todict(self):
        return {'name': self.__class__.__name__,
                'ext_i': [ext.todict() for ext in self.ext_i]}


class PointChargePotential(ExternalPotential):
    def __init__(self, charges, positions=None,
                 rc=0.2, rc2=np.inf, width=1.0):
        """Point-charge potential.

        charges: list of float
            Charges in units of `|e|`.
        positions: (N, 3)-shaped array-like of float
            Positions of charges in Angstrom.  Can be set later.
        rc: float
            Inner cutoff for Coulomb potential in Angstrom.
        rc2: float
            Outer cutoff for Coulomb potential in Angstrom.
        width: float
            Width for cutoff function for Coulomb part.

        For r < rc, 1 / r is replaced by a third order polynomial in r^2 that
        has matching value, first derivative, second derivative and integral.

        For rc2 - width < r < rc2, 1 / r is multiplied by a smooth cutoff
        function (a third order polynomium in r).

        You can also give rc a negative value.  In that case, this formula
        is used::

            (r^4 - rc^4) / (r^5 - |rc|^5)

        for all values of r - no cutoff at rc2!
        """
        self._dict = dict(name=self.__class__.__name__,
                          charges=charges, positions=positions,
                          rc=rc, rc2=rc2, width=width)
        self.q_p = np.ascontiguousarray(charges, float)
        self.rc = rc / Bohr
        self.rc2 = rc2 / Bohr
        self.width = width / Bohr
        if positions is not None:
            self.set_positions(positions)
        else:
            self.R_pv = None

        if abs(self.q_p).max() < 1e-14:
            warnings.warn('No charges!')
        if self.rc < 0. and self.rc2 < np.inf:
            warnings.warn('Long range cutoff chosen but will not be applied\
                           for negative inner cutoff values!')

    def todict(self):
        return copy.deepcopy(self._dict)

    def __str__(self):
        return ('Point-charge potential '
                '(points: {}, cutoffs: {:.3f}, {:.3f}, {:.3f} Ang)'
                .format(len(self.q_p),
                        self.rc * Bohr,
                        (self.rc2 - self.width) * Bohr,
                        self.rc2 * Bohr))

    def set_positions(self, R_pv, com_pv=None):
        """Update positions."""
        if com_pv is not None:
            self.com_pv = np.asarray(com_pv) / Bohr
        else:
            self.com_pv = None

        self.R_pv = np.asarray(R_pv) / Bohr
        self.vext_g = None

    def _molecule_distances(self, gd):
        if self.com_pv is not None:
            return self.com_pv - gd.cell_cv.sum(0) / 2

    def calculate_potential(self, gd):
        assert gd.orthogonal
        self.vext_g = gd.zeros()

        dcom_pv = self._molecule_distances(gd)

        cgpaw.pc_potential(gd.beg_c, gd.h_cv.diagonal().copy(),
                           self.q_p, self.R_pv,
                           self.rc, self.rc2, self.width,
                           self.vext_g, dcom_pv)

    def get_forces(self, calc):
        """Calculate forces from QM charge density on point-charges."""
        dens = calc.density
        F_pv = np.zeros_like(self.R_pv)
        gd = dens.finegd
        dcom_pv = self._molecule_distances(gd)

        cgpaw.pc_potential(gd.beg_c, gd.h_cv.diagonal().copy(),
                           self.q_p, self.R_pv,
                           self.rc, self.rc2, self.width,
                           self.vext_g, dcom_pv, dens.rhot_g, F_pv)
        gd.comm.sum(F_pv)
        return F_pv * Ha / Bohr


class CDFTPotential(ExternalPotential):
    # Dummy class to make cDFT compatible with new external
    # potential class ClassName(object):
    def __init__(self, regions, constraints, n_charge_regions,
                 difference):

        self.name = 'CDFTPotential'
        self.regions = regions
        self.constraints = constraints
        self.difference = difference
        self.n_charge_regions = n_charge_regions

    def todict(self):
        return {'name': 'CDFTPotential',
                # 'regions': self.indices_i,
                'constraints': self.v_i * Ha,
                'n_charge_regions': self.n_charge_regions,
                'difference': self.difference,
                'regions': self.regions}


class StepPotentialz(ExternalPotential):
    def __init__(self, zstep, value_left=0, value_right=0):
        """Step potential in z-direction

        zstep: float
            z-value that splits space into left and right [Angstrom]
        value_left: float
            Left side (z < zstep) potentential value [eV]. Default: 0
        value_right: float
            Right side (z >= zstep) potentential value [eV]. Default: 0
       """
        self.value_left = value_left
        self.value_right = value_right
        self.name = 'StepPotentialz'
        self.zstep = zstep

    def __str__(self):
        return f'Step potentialz: {self.value_left:.3f} V to '\
               f'{self.value_right:.3f} V at z={self.zstep}'

    def calculate_potential(self, gd):
        r_vg = gd.get_grid_point_coordinates()
        self.vext_g = np.where(r_vg[2] < self.zstep / Bohr,
                               gd.zeros() + self.value_left / Ha,
                               gd.zeros() + self.value_right / Ha)

    def todict(self):
        return {'name': self.name,
                'value_left': self.value_left,
                'value_right': self.value_right,
                'zstep': self.zstep}


class PotentialCollection(ExternalPotential):
    def __init__(self, potentials):
        """Collection of external potentials to be applied

        potentials: list
            List of potentials
        """
        self.potentials = []
        for potential in potentials:
            if isinstance(potential, dict):
                potential = create_external_potential(
                    potential.pop('name'), **potential)
            self.potentials.append(potential)

    def __str__(self):
        text = 'PotentialCollection:\n'
        for pot in self.potentials:
            text += '  ' + pot.__str__() + '\n'
        return text

    def calculate_potential(self, gd):
        self.potentials[0].calculate_potential(gd)
        self.vext_g = self.potentials[0].vext_g.copy()
        for pot in self.potentials[1:]:
            pot.calculate_potential(gd)
            self.vext_g += pot.vext_g

    def todict(self):
        return {'name': 'PotentialCollection',
                'potentials': [pot.todict() for pot in self.potentials]}


def static_polarizability(atoms, strength=0.01):
    """Calculate polarizability tensor

    atoms: Atoms object
    strength: field strength in V/Ang

    Returns
    -------
    polarizability tensor:
        Unit (e^2 Angstrom^2 / eV).
        Multiply with Bohr * Ha to get (Angstrom^3)
    """
    atoms.get_potential_energy()
    calc = atoms.calc
    assert calc.parameters.external is None
    dipole_gs = calc.get_dipole_moment()

    alpha = np.zeros((3, 3))
    for c in range(3):
        axes = np.zeros(3)
        axes[c] = 1
        calc.set(external=ConstantElectricField(strength, axes))
        calc.get_potential_energy()
        alpha[c] = (calc.get_dipole_moment() - dipole_gs) / strength
    calc.set(external=None)

    return alpha.T



import pytest
from functools import partial

import numpy as np

from gpaw.response.goldstone import find_root


def parabolic_function(lambd: float, *, root: float):
    """Define a parabolic function f(λ) with a root nearby λ=1.

    Defining,

    f(λ) = aλ² + bλ + c

    we need the function to be monotonically decreasing in the range
    λ∊]0.1, 10[. With parametrization

    (a, b) = (1/4, -6)

    the minimum lies at λ = -b/(2a) = 12 and the lower root at
    λ_r = -(b+d)/(2a) = 12 - d/(2a).

    Solving for c;

    d = ⎷(b²-4ac) = 2a (12 - λ_r)

    c = b² - (12 - λ_r)²/4
    """
    assert 0.1 < root < 10.
    a = 1 / 4.
    b = -6.
    c = b**2. - (12. - root)**2. / 4.
    return a * lambd**2. + b * lambd + c


@pytest.mark.response
@pytest.mark.parametrize('target', np.linspace(0.4, 2.4, 51))
def test_find_root(target):
    fnct = partial(parabolic_function, root=target)

    def is_converged(value):
        return 0. <= value < 1e-7

    root = find_root(fnct, is_converged)
    assert root == pytest.approx(target)
    assert is_converged(fnct(root))



from __future__ import annotations
from time import ctime

from typing import TYPE_CHECKING
import numpy as np
from ase.units import Ha

from gpaw.response.integrators import Integrand, HilbertTetrahedron, Intraband
from gpaw.response.chi0_base import Chi0ComponentCalculator
from gpaw.response.pair_functions import SingleQPWDescriptor
from gpaw.response.chi0_data import Chi0DrudeData
from gpaw.response.frequencies import FrequencyGridDescriptor

if TYPE_CHECKING:
    from gpaw.response.symmetry import PWSymmetryAnalyzer


class Chi0DrudeCalculator(Chi0ComponentCalculator):
    """Class for calculating the plasma frequency contribution to Chi0,
    that is, the contribution from intraband transitions inside of metallic
    bands. This corresponds directly to the dielectric function in the Drude
    model."""

    def __init__(self, *args, **kwargs):
        # Serial block distribution
        super().__init__(*args, nblocks=1, **kwargs)

        # task: IntegralTask from gpaw.response.integrators
        # wd: FrequencyDescriptor from gpaw.response.frequencies
        self.task, self.wd = self.construct_integral_task_and_wd()

    def calculate(self, wd, rate) -> Chi0DrudeData:
        """Calculate the Drude dielectric response.

        Parameters
        ----------
        wd : FrequencyDescriptor from gpaw.response.frequencies
            Frequencies to evaluate the reponse function at.
        rate : float
            Plasma frequency decay rate (in eV), corresponding to the
            imaginary part of the complex frequency.
        """
        self.print_info(wd, rate)

        chi0_drude = Chi0DrudeData.from_frequency_descriptor(wd, rate)
        self._calculate(chi0_drude)

        return chi0_drude

    def _calculate(self, chi0_drude: Chi0DrudeData):
        """In-place calculation of the Drude dielectric response function,
        based on the free-space plasma frequency of the intraband transitions.
        """
        # Create a dummy plane-wave descriptor. We need this for the symmetry
        # analysis -> see discussion in gpaw.response.jdos

        # gs: ResponseGroundStateAdapter from gpaw.response.groundstate
        # gd: GridDescriptor from gpaw.grid_descriptor
        qpd = SingleQPWDescriptor.from_q([0., 0., 0.],
                                         ecut=1e-3, gd=self.gs.gd)

        # domain: Domain from from gpaw.response.integrators
        # analyzer: PWSymmetryAnalyzer from gpaw.response.symmetry
        domain, analyzer, prefactor = self.get_integration_domain(
            qpd, spins=range(self.gs.nspins))

        # The plasma frequency integral is special in the way that only
        # the spectral part is needed
        integrand = PlasmaFrequencyIntegrand(self, qpd, analyzer)

        # Integrate using temporary array
        tmp_plasmafreq_wvv = np.zeros((1,) + chi0_drude.vv_shape, complex)

        # integrator: Integrator from gpaw.response.integrators (or child)
        self.integrator.integrate(task=self.task,
                                  domain=domain,  # Integration domain
                                  integrand=integrand,
                                  wd=self.wd,
                                  out_wxx=tmp_plasmafreq_wvv)  # Output array
        tmp_plasmafreq_wvv *= prefactor

        # Store the plasma frequency itself and print it for anyone to use
        plasmafreq_vv = tmp_plasmafreq_wvv[0].copy()
        analyzer.symmetrize_wvv(plasmafreq_vv[np.newaxis])
        chi0_drude.plasmafreq_vv += 4 * np.pi * plasmafreq_vv
        self.context.print('Plasma frequency:', flush=False)
        self.context.print((chi0_drude.plasmafreq_vv**0.5 * Ha).round(2))

        # Calculate the Drude dielectric response function from the
        # free-space plasma frequency
        # χ_D(ω+iη) = ω_p^2 / (ω+iη)^2

        # zd: ComplexFrequencyDescriptor from gpaw.response.frequencies
        assert chi0_drude.zd.upper_half_plane
        chi0_drude.chi_Zvv += plasmafreq_vv[np.newaxis] \
            / chi0_drude.zd.hz_z[:, np.newaxis, np.newaxis]**2

    def construct_integral_task_and_wd(self):
        if self.integrationmode == 'tetrahedron integration':
            # Calculate intraband transitions at T=0
            fermi_level = self.gs.fermi_level
            wd = FrequencyGridDescriptor([-fermi_level])
            task = HilbertTetrahedron(self.integrator.blockcomm)
        else:
            task = Intraband()

            # We want to pass None for frequency descriptor, but
            # if that goes wrong we'll get TypeError which is unhelpful.
            # This dummy class will give us error messages that allow finding
            # this spot in the code.
            class NotAFrequencyDescriptor:
                pass

            wd = NotAFrequencyDescriptor()
        return task, wd

    def print_info(self, wd, rate):
        isl = ['',
               f'{ctime()}',
               'Calculating drude chi0 with:',
               f'    Number of frequency points:{len(wd)}',
               f'    Plasma frequency decay rate: {rate} eV',
               '',
               self.get_gs_info_string(tab='    ')]

        # context: ResponseContext from gpaw.response.context
        self.context.print('\n'.join(isl))


class PlasmaFrequencyIntegrand(Integrand):
    def __init__(self, chi0drudecalc: Chi0DrudeCalculator,
                 qpd: SingleQPWDescriptor,
                 analyzer: PWSymmetryAnalyzer):
        self._drude = chi0drudecalc
        self.qpd = qpd
        self.analyzer = analyzer

    def _band_summation(self):
        # Intraband response needs only integrate partially unoccupied bands.
        # gs: ResponseGroundStateAdapter from gpaw.response.groundstate
        return self._drude.gs.nocc1, self._drude.gs.nocc2

    def matrix_element(self, point):
        """NB: In dire need of documentation! XXX."""
        k_v = point.kpt_c  # XXX _v vs _c discrepancy
        n1, n2 = self._band_summation()
        k_c = np.dot(self.qpd.gd.cell_cv, k_v) / (2 * np.pi)

        # kptpair_factory: KPointPairFactory from gpaw.response.pair
        kptpair_factory = self._drude.kptpair_factory

        K0 = kptpair_factory.gs.kpoints.kptfinder.find(k_c)  # XXX

        kpt1 = kptpair_factory.get_k_point(point.spin, K0, n1, n2)
        n_n = range(n1, n2)

        pair_calc = kptpair_factory.pair_calculator(
            blockcomm=self._drude.blockcomm)
        vel_nv = pair_calc.intraband_pair_density(kpt1, n_n)

        if self._drude.integrationmode is None:
            f_n = kpt1.f_n
            width = self._drude.gs.get_occupations_width()
            if width > 1e-15:
                dfde_n = - 1. / width * (f_n - f_n**2.0)
            else:
                dfde_n = np.zeros_like(f_n)
            vel_nv *= np.sqrt(-dfde_n[:, np.newaxis])
            weight = np.sqrt(self.analyzer.get_kpoint_weight(k_c) /
                             self.analyzer.how_many_symmetries())
            vel_nv *= weight

        return vel_nv

    def eigenvalues(self, point):
        """A function that can return the intraband eigenvalues.

        A method describing the integrand of
        the response function which gives an output that
        is compatible with the gpaw k-point integration
        routines."""
        n1, n2 = self._band_summation()
        # gs: ResponseGroundStateAdapter from gpaw.response.groundstate
        gs = self._drude.gs
        # kd: KPointDescriptor object from gpaw.kpt_descriptor
        kd = gs.kd
        k_v = point.kpt_c  # XXX v/c discrepancy
        # gd: GridDescriptor from gpaw.grid_descriptor
        k_c = np.dot(self.qpd.gd.cell_cv, k_v) / (2 * np.pi)
        K1 = gs.kpoints.kptfinder.find(k_c)
        ik = kd.bz2ibz_k[K1]
        kpt1 = gs.kpt_qs[ik][point.spin]
        assert gs.kd.comm.size == 1

        return kpt1.eps_n[n1:n2]



"""
Helper class for LCAOETDM.

Handles orbital initialization, setting reference orbitals, applying the
unitary transformation, calculating the gradient, and getting canonical
representation.
"""

import numpy as np
from gpaw.lcao.eigensolver import DirectLCAO
from gpaw.directmin.functional.lcao import get_functional
from gpaw.utilities.tools import tri2full


class ETDMHelperLCAO(DirectLCAO):

    def __init__(self, wfs, dens, ham, nkpts, functional, diagonalizer=None,
                 orthonormalization='gramschmidt',
                 need_init_orbs=True):

        super(ETDMHelperLCAO, self).__init__(diagonalizer)
        super(ETDMHelperLCAO, self).initialize(wfs.gd, wfs.dtype,
                                               wfs.setups.nao, wfs.ksl)
        self.orthonormalization = orthonormalization
        self.need_init_orbs = need_init_orbs
        self.nkpts = nkpts
        self.func = get_functional(functional, wfs, dens, ham)
        self.reference_orbitals = {}
        self.initialize_orbitals(wfs, ham)

    def __repr__(self):
        pass

    def set_reference_orbitals(self, wfs, n_dim):
        for kpt in wfs.kpt_u:
            u = self.kpointval(kpt)
            self.reference_orbitals[u] = np.copy(kpt.C_nM[:n_dim[u]])

    def appy_transformation_kpt(self, wfs, u_mat, kpt, c_ref=None,
                                broadcast=True,
                                update_proj=True):
        """
        If c_ref are not provided then
        kpt.C_nM <- u_mat kpt.C_nM
        otherwise kpt.C_nM <- u_mat c_ref
        """

        dimens1 = u_mat.shape[1]
        dimens2 = u_mat.shape[0]

        if c_ref is None:
            kpt.C_nM[:dimens2] = u_mat @ kpt.C_nM[:dimens1]
        else:
            kpt.C_nM[:dimens2] = u_mat @ c_ref[:dimens1]

        if broadcast:
            with wfs.timer('Broadcast coefficients'):
                wfs.gd.comm.broadcast(kpt.C_nM, 0)
        if update_proj:
            with wfs.timer('Calculate projections'):
                wfs.atomic_correction.calculate_projections(wfs, kpt)

    def initialize_orbitals(self, wfs, ham):
        """
        If it is the first use of the scf then initialize
        coefficient matrix using eigensolver
        """

        orthname = self.orthonormalization
        need_canon_coef = \
            (not wfs.coefficients_read_from_file and self.need_init_orbs)
        if need_canon_coef or orthname == 'diag':
            super(ETDMHelperLCAO, self).iterate(ham, wfs)
        else:
            wfs.orthonormalize(type=orthname)
        wfs.coefficients_read_from_file = False
        self.need_init_orbs = False

    def calc_grad(self, wfs, ham, kpt, evecs, evals, matrix_exp,
                  representation, ind_up, constraints):
        """
        Gradient w.r.t. skew-Hermitian matrices
        """

        h_mm = self.calculate_hamiltonian_matrix(ham, wfs, kpt)
        # make matrix hermitian
        tri2full(h_mm)
        # calc gradient and eigenstate error
        g_mat, error = self.func.get_gradients(
            h_mm, kpt.C_nM, kpt.f_n, evecs, evals,
            kpt, wfs, wfs.timer, matrix_exp,
            representation, ind_up, constraints)

        return g_mat, error

    def update_to_canonical_orbitals(self, wfs, ham, kpt,
                                     update_ref_orbs_canonical, restart):
        """
        Choose canonical orbitals
        """

        h_mm = self.calculate_hamiltonian_matrix(ham, wfs, kpt)
        tri2full(h_mm)

        if self.func.name == 'ks':
            if update_ref_orbs_canonical or restart:
                # Diagonalize entire Hamiltonian matrix
                with wfs.timer('Diagonalize and rotate'):
                    kpt.C_nM, kpt.eps_n = rotate_subspace(h_mm, kpt.C_nM)
            else:
                # Diagonalize equally occupied subspaces separately
                f_unique = np.unique(kpt.f_n)
                for f in f_unique:
                    with wfs.timer('Diagonalize and rotate'):
                        kpt.C_nM[kpt.f_n == f, :], kpt.eps_n[kpt.f_n == f] = \
                            rotate_subspace(h_mm, kpt.C_nM[kpt.f_n == f, :])
        elif self.func.name == 'PZ-SIC':
            self.func.get_lagrange_matrices(
                h_mm, kpt.C_nM, kpt.f_n, kpt, wfs,
                update_eigenvalues=True)

        with wfs.timer('Calculate projections'):
            self.update_projections(wfs, kpt)

    def sort_orbitals(self, wfs, kpt, ind):
        """
        sort orbitals according to indices stored in ind
        """
        kpt.C_nM[np.arange(len(ind)), :] = kpt.C_nM[ind, :]
        self.update_projections(wfs, kpt)

    def update_projections(self, wfs, kpt):
        """
        calculate projections kpt.P_ani
        """

        wfs.atomic_correction.calculate_projections(wfs, kpt)

    def orbital_energies(self, wfs, ham, kpt):
        """
        diagonal elements of hamiltonian matrix in orbital representation
        """

        if self.func.name == 'ks':
            h_mm = self.calculate_hamiltonian_matrix(ham, wfs, kpt)
            tri2full(h_mm)
            # you probably need only diagonal terms?
            # wouldn't "for" be faster?
            h_mm = kpt.C_nM.conj() @ h_mm.conj() @ kpt.C_nM.T
            energies = h_mm.diagonal().real.copy()
        elif self.func.name == 'PZ-SIC':
            energies = self.func.lagr_diag_s[wfs.eigensolver.kpointval(kpt)]

        return energies

    def kpointval(self, kpt):
        return self.nkpts * kpt.s + kpt.q


def rotate_subspace(h_mm, c_nm):
    """
    choose canonical orbitals
    """
    l_nn = c_nm @ h_mm @ c_nm.conj().T
    # check if diagonal then don't rotate? it could save a bit of time
    eps, w = np.linalg.eigh(l_nn)
    return w.T.conj() @ c_nm, eps



import pytest
from ase import Atoms
from ase.dft.kpoints import monkhorst_pack
from gpaw import GPAW, PW, mpi
from gpaw.pipekmezey.pipek_mezey_wannier import PipekMezey


@pytest.mark.pipekmezey
def test_pipekmezey_chain(in_tmp_dir):

    atoms = Atoms('H4',
                  positions=[[0, 0, 0],
                             [0, 0, 0.74],
                             [0, 0, 1.48],
                             [0, 0, 2.22]])
    atoms.cell = [10, 10, 2.98]
    atoms.center()
    atoms.pbc = (False, False, True)

    kpts = monkhorst_pack((1, 1, 4))

    calc = GPAW(mode=PW(200),
                h=0.24,
                kpts=kpts,
                convergence={'density': 1e-4,
                             'eigenstates': 1e-4},
                symmetry={'point_group': False,
                          'time_reversal': False},
                parallel={'domain': mpi.world.size})

    calc.atoms = atoms
    calc.calculate()

    PM = PipekMezey(calc=calc, seed=42)
    PM.localize()

    P = PM.get_function_value()

    assert P == pytest.approx(0.6803, abs=0.0001)



import pytest

from gpaw import GPAW, LCAO
from gpaw.directmin.etdm_lcao import LCAOETDM
from ase import Atoms
import numpy as np


@pytest.mark.later
@pytest.mark.sic
def test_lcaosic(in_tmp_dir):
    """
    Test Perdew-Zunger Self-Interaction
    Correction  in LCAO mode using ETDM
    :param in_tmp_dir:
    :return:
    """

    # Water molecule:
    d = 0.9575
    t = np.pi / 180 * 104.51
    H2O = Atoms('OH2',
                positions=[(0, 0, 0),
                           (d, 0, 0),
                           (d * np.cos(t), d * np.sin(t), 0)])
    H2O.center(vacuum=4.0)

    calc = GPAW(mode=LCAO(force_complex_dtype=True),
                h=0.22,
                occupations={'name': 'fixed-uniform'},
                eigensolver=LCAOETDM(localizationtype='PM_PZ',
                                     localizationseed=42,
                                     functional={'name': 'PZ-SIC',
                                                 'scaling_factor':
                                                     (0.5, 0.5)}),
                convergence={'eigenstates': 1e-4},
                mixer={'backend': 'no-mixing'},
                nbands='nao',
                symmetry='off'
                )
    H2O.calc = calc
    e = H2O.get_potential_energy()
    f = H2O.get_forces()

    assert e == pytest.approx(-12.16352, abs=1e-3)

    f2 = np.array([[-4.21747862, -4.63118948, 0.00303988],
                   [5.66636141, -0.51037693, -0.00049136],
                   [-1.96478031, 5.4043045, -0.0006107]])
    assert f2 == pytest.approx(f, abs=0.1)

    numeric = False
    if numeric:
        from ase.calculators.test import numeric_force
        f_num = np.array([[numeric_force(H2O, a, i)
                          for i in range(3)]
                         for a in range(len(H2O))])
        print('Numerical forces')
        print(f_num)
        print(f - f_num, np.abs(f - f_num).max())

    calc.write('h2o.gpw', mode='all')
    from gpaw import restart
    H2O, calc = restart('h2o.gpw', txt='-')
    H2O.positions += 1.0e-6
    f3 = H2O.get_forces()
    niter = calc.get_number_of_iterations()
    assert niter == pytest.approx(4, abs=3)
    assert f2 == pytest.approx(f3, abs=0.1)



from gpaw.grid_descriptor import GridDescriptor
from gpaw.poisson import FDPoissonSolver
import pytest
import numpy as np
from gpaw.mpi import size


@pytest.mark.gpu
@pytest.mark.skipif(size == 8, reason='Fails at the moment for size=8')
def test_poisson(gpu):
    import cupy
    phis = []
    for xp in [np, cupy]:
        lat = 8.0
        gd = GridDescriptor((8, 6, 8), (lat, lat, lat),
                            pbc_c=[False, False, False])
        # Use Gaussian as input
        x, y, z = gd.get_grid_point_coordinates()
        x, y, z = xp.asarray(x), xp.asarray(y), xp.asarray(z)
        sigma = 1.5
        mu = lat / 2.0

        rho = gd.zeros(xp=xp)
        rho[:] = xp.exp(
            -((x - mu)**2 + (y - mu)**2 + (z - mu)**2) / (2.0 * sigma))
        charge = gd.integrate(rho)
        rho -= charge * gd.dv

        phi = gd.zeros(xp=xp)

        poisson = FDPoissonSolver(xp=xp)
        poisson.set_grid_descriptor(gd)
        poisson.solve(phi, rho)
        phis.append(phi)
    cupy.allclose(phis[0], phis[1], rtol=1e-10)
